from CIDAN.DataHandler import DataHandler
from PySide2.QtWidgets import *
# from PySide2.QtGui import *
from PySide2 import QtCore
# def loadImageWrapper(main_widget):
#     def loadImage():
#         path_to_file = createFileDialog("~/Desktop", forOpen=True, fmt='', isFolder=0)
#         path_to_save_dir = createFileDialog("~/Desktop", forOpen=False, fmt='',
#                                             isFolder=1)
#         if hasattr(main_widget, "data_handler"):
#             main_widget.data_handler.__del__()
#         main_widget.data_handler = DataHandler(data_path=path_to_file,
#                                                save_dir_path=path_to_save_dir,
#                                                save_dir_already_created=False)
#         main_widget.image_view.setImage(main_widget.data_handler.calculate_filters())
#
#     return loadImage

def createFileDialog(directory='', forOpen=True, fmt='', isFolder=0):
    directory="/Users/sschickler/Documents/LSSC-python/input_images"
    options = QFileDialog.Options()
    options |= QFileDialog.DontUseNativeDialog
    options |= QFileDialog.DontUseCustomDirectoryIcons
    dialog = QFileDialog()
    dialog.setOptions(options)

    dialog.setFilter(dialog.filter() | QtCore.QDir.Hidden)

    # ARE WE TALKING ABOUT FILES OR FOLDERS
    if isFolder == 1:
        dialog.setFileMode(QFileDialog.DirectoryOnly)
    if isFolder == 2:
        dialog.setFileMode(QFileDialog.AnyFile)
    # OPENING OR SAVING
    dialog.setAcceptMode(QFileDialog.AcceptOpen) if forOpen else dialog.setAcceptMode(QFileDialog.AcceptSave)

    # SET FORMAT, IF SPECIFIED
    if fmt != '' and isFolder is False:
        dialog.setDefaultSuffix(fmt)
        dialog.setNameFilters([f'{fmt} (*.{fmt})'])

    # SET THE STARTING DIRECTORY
    if directory != '':
        dialog.setDirectory(str(directory))
    # else:
    #     dialog.setDirectory(str(ROOT_DIR))


    if dialog.exec_() == QDialog.Accepted:
        path = dialog.selectedFiles()[0]  # returns a list
        return path
    else:
        return ''

def load_new_dataset(main_widget, file_input, save_dir_input):
    # TODO add error handling here to ensure valid inputs
    file_path = file_input.current_state()
    save_dir_path = save_dir_input.current_state()
    # if hasattr(main_widget, "data_handler"):
    #     main_widget.data_handler.__del__()

    main_widget.data_handler = DataHandler(data_path=file_path,
                                               save_dir_path=save_dir_path,
                                               save_dir_already_created=False)
    main_widget.init_w_data()

def load_prev_session(main_widget, save_dir_input):
    # TODO add error handling here to ensure valid inputs
    save_dir_path = save_dir_input.current_state()
    main_widget.data_handler = DataHandler(data_path="",
                                           save_dir_path=save_dir_path,
                                           save_dir_already_created=True)
    main_widget.init_w_data()
def export_timetraces(main_widget):
    createFileDialog(forOpen=False,isFolder=2)



import os
import numpy as np
import tifffile
import matplotlib.pyplot as plt
from scipy import ndimage
from typing import Union, Any, List, Optional, cast, Tuple, Dict

# from IPython.display import display, Image

def load_filter_tif_stack(*, path, filter: bool, median_filter: bool,
                          median_filter_size: Tuple[int, int,int],
                          z_score: bool, slice_stack: bool,
                          slice_every, slice_start: int):
    """
    This function reads a tiff stack file
    Parameters
    ----------
    path The path to a single tif stack or a list of paths to many tiff stacks


    Returns
    -------
    a 3D numpy array with the tiff files together

    """

    if type(path) == list or os.path.isdir(path):
        volumes = []
        paths = path if type(path) == list else sorted(os.listdir(path))

        for num, x in enumerate(paths):
            file_path = os.path.join(path, x)
            image = tifffile.imread(file_path)
            if slice_stack:
                image = image[slice_start::slice_every, :, :]
            if filter:
                image = filter_stack(stack=image, median_filter=median_filter,
                                     median_filter_size=median_filter_size,
                                     z_score=z_score)
            volumes.append(image)
            print("Loading: "+x)

        image = np.vstack(volumes)
        del volumes
        return image
    if os.path.isfile(path):
        # return ScanImageTiffReader(path).data()
        image = tifffile.imread(path)

        if slice_stack:
            image = image[slice_start::slice_every,:,:]
        if filter:
            image = filter_stack(stack=image, median_filter=median_filter,
                                 median_filter_size=median_filter_size, z_score=z_score)
        return image
    raise Exception("Invalid Input folders not allowed currently ")
    # vol=ScanImageTiffReader(file_path).data()

def reshape_to_2d_over_time(volume):
    """
    Takes a 3d numpy volume with dim 1 and 2 being x, y and dim 0 being time
    and return array compressing dim 1 and 2 into 1 dimension

    Parameters
    ----------
    volume input 3d numpy volume

    Returns
    -------
    a 2d volume with the 0 dim being pixel number and 1 being time value

    """
    return np.transpose(np.reshape(volume, (volume.shape[0],-1), order="C"))
def save_image(volume: np.ndarray, name: str, directory: str, shape: tuple, number_save: int):
    """
    Function to save images from a 3d volume
    Parameters
    ----------
    volume
    name
    directory
    shape shape to reshape volume into
    number_save number of slices to take and save

    Returns
    -------
    None
    """
    for x in range(number_save):
        fig, ax = plt.subplots()
        imgplot = ax.imshow(np.reshape(volume,
                       shape)[shape[0]//number_save*x])
        fig.savefig(os.path.join(directory,name+"_"+str(x)))

        # img = Image.fromarray(
        #     np.reshape(volume,
        #                shape)[shape[0]/number_save*x] * 255).convert('L')
        # img.save(
        #     os.path.join(directory,name+"_"+str(x)))
def filter_stack(*,stack: np.ndarray, median_filter: bool,
                 median_filter_size: Tuple[int, int, int],
                 z_score: bool,):

    if z_score:
        stack_t  = np.transpose(stack, (1, 2, 0))
        shape = (1,stack_t.shape[1],stack_t.shape[2])
        std = np.std(stack_t,axis=0).reshape(shape)
        mean = np.mean(stack_t,axis=0).reshape(shape)
        stack_t = (stack_t-mean)/std
        stack = np.transpose(stack_t, (2, 0, 1))
    if median_filter:
        stack = ndimage.median_filter(stack, median_filter_size)
    return stack

def pixel_num_to_2d_cord(pixel_list, volume_shape):
    """
    Takes a list of pixels and converts it to a list of cords for each pixel

    Parameters
    ----------
    pixel_list : np.ndarray
        list of pixel numbers
    volume_shape : tuple
        shape of the original dataset

    Returns
    -------
    a list of tuples where each is the cordinates for each pixel

    """
    def convert_one(num):
        return num%volume_shape[2], num//volume_shape[2]

    return np.apply_along_axis(convert_one,axis=0, arr=pixel_list)


from PySide2.QtWidgets import *
from CIDAN.Input import *
class SettingBlockModule(QFrame):
    def __init__(self, name, input_list):
        super().__init__()
        self.name = name
        self.input_list = input_list
        self.layout = QVBoxLayout()
        #Type options are int, float, bool, and str
        for input in input_list:
            self.layout.addWidget(input)

        self.setLayout(self.layout)

def filter_setting_block(main_widget):
    data_handler = main_widget.data_handler
    return SettingBlockModule("Filter Settings", [BoolInput(display_name="Median filter:",
                                                           program_name="median_filter",
                                                           on_change_function=lambda x,y:data_handler.change_filter_param(x,y),
                                                           default_val=data_handler.filter_params["median_filter"],
                                                           tool_tip="Whether to apply a median filter to each timestep",
                                                           ),
                                                  IntInput(
                                                      display_name="Median filter size:",
                                                      program_name="median_filter_size",
                                                      on_change_function=lambda x, y:data_handler.change_filter_param(x,y),
                                                      default_val=
                                                      data_handler.filter_params[
                                                          "median_filter_size"],
                                                      tool_tip="The size of the median filter for each timestep",
                                                      min=1,max=50,step=1),
                                                  BoolInput(
                                                      display_name="Z-score:",
                                                      program_name="z_score",
                                                      on_change_function=lambda x,y:data_handler.change_filter_param(x,y),
                                                      default_val=
                                                      data_handler.filter_params["z_score"],
                                                      tool_tip="Whether to apply a z-score for each pixel across all the timesteps",
                                                      )
                                                  ])

def dataset_setting_block(main_widget):
    data_handler = main_widget.data_handler
    return SettingBlockModule("Dataset Settings",[BoolInput(display_name="Slice stack:",
                                                   program_name="slice_stack",
                                                    on_change_function=lambda x,y:data_handler.change_dataset_param(x,y),
                                                    default_val=data_handler.dataset_params["slice_stack"],
                                                    tool_tip="Used to Select only "
                                                             "certain timepoints from a dataset", display_tool_tip=True),
                                                  IntInput(display_name="Slice every:",
                                                           program_name="slice_every",
                                                           on_change_function=lambda x,y:data_handler.change_dataset_param(x,y),
                                                           default_val=data_handler.dataset_params["slice_every"],
                                                           tool_tip="Select every x timesteps",
                                                           min=1, max=100,step=1),
                                                  IntInput(display_name="Slice start:",
                                                           program_name="slice_start",
                                                           on_change_function=lambda x,y:data_handler.change_dataset_param(x,y),
                                                           tool_tip="Start selecting on x timestep",
                                                           default_val=data_handler.dataset_params["slice_start"],
                                                           min=0, max=1000,step=1)
                                                  ])

def multiprocessing_settings_block(main_widget):
    data_handler = main_widget.data_handler
    return SettingBlockModule("Multiprocessing Settings", [IntInput(display_name="Number of timesteps:",
                                                           program_name="total_num_time_steps",
                                                           on_change_function=lambda x,y:data_handler.change_box_param(x,y),
                                                           default_val=data_handler.box_params["total_num_time_steps"],
                                                           tool_tip="Number of timesteps to break the processing into",
                                                           min=1, max=1000,step=1),
                                                           IntInput(
                                                               display_name="Number of spatial boxes:",
                                                               program_name="total_num_spatial_boxes",
                                                               on_change_function=lambda x,y: data_handler.change_box_param(
                                                                   x, y),
                                                               default_val=
                                                               data_handler.box_params[
                                                                   "total_num_spatial_boxes"],
                                                               tool_tip="Number of boxes to break the calculation into, please make sure it is a sqaure",
                                                               min=1, max=1000, step=1)
                                                           ,
                                                           IntInput(
                                                              display_name="Spatial overlap:",
                                                              program_name="spatial_overlap",
                                                              on_change_function=lambda x, y: data_handler.change_box_param(x, y),
                                                              default_val=
                                                              data_handler.box_params[
                                                                  "spatial_overlap"],
                                                              tool_tip="Number of pixels to overlap each box",
                                                              min=1, max=1000, step=1)])

def roi_extraction_settings_block(main_widget):
    # TODO fix input into spatial box number
    data_handler = main_widget.data_handler
    return SettingBlockModule("ROI Extraction Settings",
                              [
                               IntInput(display_name="Max Number of ROIs per spatial box:",
                                        program_name="num_rois",
                                        on_change_function=lambda x,
                                                                  y: data_handler.change_roi_extraction_param(
                                            x, y),
                                        default_val=data_handler.roi_extraction_params[
                                            "num_rois"],
                                        tool_tip="Max number of ROIs to select for each spatial box",
                                        min=1, max=1000, step=1),
                               IntInput(
                                   display_name="ROI size minimum:",
                                   program_name="roi_size_min",
                                   on_change_function=lambda x,
                                                             y: data_handler.change_roi_extraction_param(
                                       x, y),
                                   default_val=
                                   data_handler.roi_extraction_params[
                                       "roi_size_min"],
                                   tool_tip="Minimum size in pixels for a region of interest",
                                   min=1, max=1000, step=1)
                                  ,
                               IntInput(
                                   display_name="ROI size maximum:",
                                   program_name="roi_size_max",
                                   on_change_function=lambda x,
                                                             y: data_handler.change_roi_extraction_param(
                                       x, y),
                                   default_val=
                                   data_handler.roi_extraction_params[
                                       "roi_size_max"],
                                   tool_tip="Maximum size in pixels for a region of interest",
                                   min=1, max=100000, step=1),
                               BoolInput(
                                   display_name="Run refinement step:",
                                   program_name="refinement",
                                   on_change_function=lambda x,
                                                             y: data_handler.change_roi_extraction_param(
                                       x, y),
                                   default_val=
                                   data_handler.roi_extraction_params[
                                       "refinement"],
                                   tool_tip="Whether to run the refinement step, greatly improves results"),
                               FloatInput(
                                   display_name="Merge temporal coefficient:",
                                   program_name="merge_temporal_coef",
                                   on_change_function=lambda x,
                                                             y: data_handler.change_roi_extraction_param(
                                       x, y),
                                   default_val=
                                   data_handler.roi_extraction_params[
                                       "merge_temporal_coef"],
                                   tool_tip="The coefficient that determines if two overlapping regions are merged based on their temporal correlation",
                                   min=0, max=1, step=.01),
                               BoolInput(
                                   display_name="Fill holes:",
                                   program_name="fill_holes",
                                   on_change_function=lambda x,
                                                             y: data_handler.change_roi_extraction_param(
                                       x, y),
                                   default_val=
                                   data_handler.roi_extraction_params[
                                       "fill_holes"],
                                   tool_tip="Whether to fill holes in each cluster"),

                               ]
                                            )
def roi_advanced_settings_block(main_widget):
    # TODO fix input into spatial box number
    data_handler = main_widget.data_handler
    return SettingBlockModule("Advanced Settings",
                              [IntInput(display_name="Number of eigen vectors:",
                                        program_name="num_eig",
                                        on_change_function=lambda x,
                                                                  y: data_handler.change_eigen_param(
                                            x, y),
                                        default_val=data_handler.eigen_params[
                                            "num_eig"],
                                        tool_tip="Number of eigen vectors to generate",
                                        min=1, max=1000, step=1),


                               ]
                                            )
# TODO add fill holes/number rois per spatial box to inputs


from CIDAN.Tab import Tab
from PySide2.QtWidgets import *
from CIDAN.DataHandlerWrapper import PreprocessThread
from CIDAN.SettingsModule import preprocessing_settings


class PreprocessingTab(Tab):
    """Class controlling the Preprocessing tab, inherits from Tab


    Attributes
    ----------
    main_widget : MainWidget
        A reference to the main widget
    data_handler : DataHandler
        A reference to the main DataHandler of MainWidget

    """

    def __init__(self, main_widget):
        self.main_widget = main_widget
        self.data_handler = self.main_widget.data_handler

        # This part initializes the button to process the data
        process_button = QPushButton()
        process_button.setText("Apply Settings")
        thread = PreprocessThread(main_widget, process_button)
        main_widget.thread_list.append(thread) # Appends the thread to the main
        # widget thread list
        process_button.clicked.connect(lambda: thread.runThread())
        # This assumes that the data is already loaded in
        self.main_widget.preprocess_image_view.setImage(
            self.data_handler.calculate_filters())

        # Section that creates all the buttons to change which image is displayed
        image_buttons = QWidget()
        image_buttons_layout = QHBoxLayout()
        image_buttons.setLayout(image_buttons_layout)
        max_image_button = QPushButton()
        max_image_button.setText("Max Image")
        max_image_button.clicked.connect(
            lambda: self.main_widget.preprocess_image_view.setImage(
                self.data_handler.max_image))
        stack_button = QPushButton()
        stack_button.setText("Filtered Stack")
        stack_button.clicked.connect(
            lambda: self.main_widget.preprocess_image_view.setImage(
                self.data_handler.dataset_filtered))
        orig_stack_button = QPushButton()
        orig_stack_button.setText("Original Stack")
        orig_stack_button.clicked.connect(
            lambda: self.main_widget.preprocess_image_view.setImage(
                self.data_handler.dataset))
        mean_image_button = QPushButton()
        mean_image_button.setText("Mean Image")
        mean_image_button.clicked.connect(
            lambda: self.main_widget.preprocess_image_view.setImage(
                self.data_handler.mean_image))
        image_buttons_layout.addWidget(orig_stack_button)
        image_buttons_layout.addWidget(stack_button)
        image_buttons_layout.addWidget(max_image_button)
        image_buttons_layout.addWidget(mean_image_button)

        # Initialize the tab with the necessary columns
        super().__init__("Preprocessing", column_1=[preprocessing_settings(main_widget),
                                                    process_button], column_2=[
            main_widget.preprocess_image_view, image_buttons])



from CIDAN import MainWindow
from PySide2.QtWidgets import QApplication
import sys
import logging

LEVELS = {'debug': logging.DEBUG,
          'info': logging.INFO,
          'warning': logging.WARNING,
          'error': logging.ERROR,
          'critical': logging.CRITICAL}

if len(sys.argv) > 1:
    LOG_FILENAME = 'log.out'
    level_name = sys.argv[1]
    level = LEVELS.get(level_name, logging.NOTSET)
    logging.basicConfig(filename=LOG_FILENAME,level=level)
    logger = logging.getLogger("CIDAN")
    logger.debug("Program started")


app = QApplication([])
app.setApplicationName("CIDAN")
widget = MainWindow.MainWindow()

sys.exit(app.exec_())


from CIDAN.Tab import Tab
from CIDAN.Input import FileInput
from PySide2.QtWidgets import *
from CIDAN.fileHandling import *
class FileOpenTab(Tab):
    def __init__(self,main_widget):
        # TODO Make this less ugly can reorganize code
        dataset_file_input = FileInput("Dataset File:", "", "","","Select a file to load in", isFolder=2, forOpen=True)
        dataset_folder_input = FileInput("Dataset Folder:", "", "","","Select a folder to load in", isFolder=1, forOpen=True)
        save_dir_new_file = FileInput("Save Directory Location:", "", "","","Select a place to save outputs", isFolder=1, forOpen=False)
        save_dir_new_folder = FileInput("Save Directory Location:", "", "","","Select a place to save outputs", isFolder=1, forOpen=False)

        save_dir_load = FileInput("Previous Session Location:", "", "", "",
                      "Select the save directory for a previous session", isFolder=1, forOpen=True)
        file_open_button = QPushButton()
        file_open_button.setText("Load")
        file_open_button.clicked.connect(lambda: load_new_dataset(main_widget,dataset_file_input,save_dir_new_file))
        folder_open_button = QPushButton()
        folder_open_button.setText("Load")
        folder_open_button.clicked.connect(
            lambda: load_new_dataset(main_widget, dataset_folder_input, save_dir_new_folder))
        prev_session_open_button = QPushButton()
        prev_session_open_button.setText("Load")
        prev_session_open_button.clicked.connect(
            lambda: load_prev_session(main_widget, save_dir_load))
        file_open = Tab("File Open", column_2=[], column_2_display=False,column_1=[dataset_file_input,save_dir_new_file, file_open_button])
        folder_open = Tab("Folder Open", column_2=[], column_2_display=False,column_1=[dataset_folder_input,save_dir_new_folder,folder_open_button]
                                               )
        prev_session_open = Tab("Previous Session Open", column_2=[], column_2_display=False,column_1=[save_dir_load,prev_session_open_button])
        self.tab_selector = QTabWidget()
        self.tab_selector.addTab(file_open, file_open.name)
        self.tab_selector.addTab(folder_open, folder_open.name)
        self.tab_selector.addTab(prev_session_open, prev_session_open.name)

        super().__init__("FileOpenTab", column_1=[self.tab_selector], column_2=[], column_2_display=False)


from dask import delayed
import numpy as np
import logging
logger1 = logging.getLogger("CIDAN.LSSC.SpatialBox")
class SpatialBox:
    def __init__(self, box_num: int, total_boxes: int, image_shape: int,
                 spatial_overlap: int):
        logger1.debug(
            "Spatial Box creation inputs: box num {0}, total boxes {1}, image shape {2}, spatial overlap {3}".format(box_num, total_boxes,image_shape,spatial_overlap))
        # TODO implement spatial overlap
        self.box_num = box_num
        self.total_boxes = total_boxes
        self.image_shape = image_shape
        self.boxes_per_row = int(total_boxes ** .5)
        self.spatial_overlap = spatial_overlap
        self.y_box_num = box_num // self.boxes_per_row
        self.x_box_num = box_num - (self.y_box_num * self.boxes_per_row)

        self.box_cord_1 = [((image_shape[1] // self.boxes_per_row) *
                           self.x_box_num)-spatial_overlap,
                           (image_shape[2] // self.boxes_per_row) *
                           self.y_box_num-spatial_overlap]
        self.box_cord_2 = [(image_shape[1] // self.boxes_per_row) * (
                self.x_box_num+1)+spatial_overlap,
                           (image_shape[2] // self.boxes_per_row ) * (
                                   self.y_box_num+1)+spatial_overlap]

        self.box_cord_1[0] = 0 if self.box_cord_1[0]<0 else self.box_cord_1[0]
        self.box_cord_1[1] =  0 if self.box_cord_1[1] < 0 \
            else self.box_cord_1[1]
        self.box_cord_2[0] = image_shape[1] if self.box_cord_2[0] > image_shape[1] \
            else \
            self.box_cord_2[0]
        self.box_cord_2[1] = image_shape[2] if self.box_cord_2[1] > image_shape[2] \
            else self.box_cord_2[1]
        self.shape = (image_shape[0],self.box_cord_2[0]-self.box_cord_1[0],
                                  self.box_cord_2[
            1]-self.box_cord_1[1])
        logger1.debug(("Spatial box creation: Boxes per row {0}, y_box_num {1}, "+
                      "x_box_num"+" {2}, box cord 1 {3}, box cord 2 {4}, shape {5}"
                       ).format(self.boxes_per_row,self.y_box_num,self.x_box_num,self.box_cord_1, self.box_cord_2, self.shape))
    @delayed
    def extract_box(self, dataset):
        return dataset[:,self.box_cord_1[0]:self.box_cord_2[0], self.box_cord_1[1]:
                       self.box_cord_2[1]]
    @delayed
    def redefine_spatial_cord_2d(self, cord_list):
        return [(x+self.box_cord_1[0],y+self.box_cord_1[1]) for x,y in cord_list]
    @delayed
    def redefine_spatial_cord_1d(self, cord_list):
        box_length = self.box_cord_2[1]-self.box_cord_1[1]
        def change_1_cord(x):
            return ((x//box_length)+self.box_cord_1[0])*self.image_shape[
                2]+self.box_cord_1[1]+x%box_length
        return list(map(change_1_cord,cord_list))

    def convert_1d_to_2d(self):
        # TODO implement this
        pass
    def data_w_out_spatial_overlap(self, data):
        """

        Parameters
        ----------
        data 2d dataset

        Returns
        -------

        """
        if self.total_boxes == 1:
            return data
        x = [0,self.shape[1]]
        y = [0, self.shape[2]]
        # This uses which column each box is in to determin overlap parts, first, last,
        # any other column
        if self.box_num%self.boxes_per_row==0:
            x[1] = self.shape[1]-self.spatial_overlap
        elif self.box_num%self.boxes_per_row==self.boxes_per_row-1:
            x[0] = self.spatial_overlap
        else:
            x[0] = self.spatial_overlap
            x[1] = self.shape[1] - self.spatial_overlap
        # This uses which row each box is in to determin overlap parts, first, last,
        # any other row
        if self.box_num//self.boxes_per_row==0:
            y[1] = self.shape[2]-self.spatial_overlap
        elif self.box_num//self.boxes_per_row==self.boxes_per_row-1:
            y[0] = self.spatial_overlap
        else:
            y[0] = self.spatial_overlap
            y[1] = self.shape[2] - self.spatial_overlap
        return data[x[0]:x[1],y[0]:y[1]]


def combine_images(spatial_box_list, data_list):
    """

    Parameters
    ----------
    spatial_box_list
    data_list list of reshaped eigen vectors in the correct shape for these boxes

    Returns
    -------

    """
    # Going to go through the lists in a 2d format using number of spatial boxes is always square
    spatial_box_num = len(spatial_box_list)
    spatial_box_root= int(spatial_box_num**.5)
    data_matched = []
    for y in range(spatial_box_root):
        temp = []
        for x in range(spatial_box_root):
            current_spatial_box = spatial_box_list[y*spatial_box_root+x]
            current_data = data_list[y*spatial_box_root+x]
            temp.append(current_spatial_box.data_w_out_spatial_overlap(current_data))
        data_matched.append(np.vstack(temp))
    all_data = np.hstack(data_matched)

    return all_data
if __name__ == '__main__':
    test = SpatialBox(box_num=0,total_boxes=9, image_shape=[1,9,9],spatial_overlap=0)
    pixel_list = test.redefine_spatial_cord_1d([0,4,8]).compute()
    import numpy as np
    zeros = np.zeros((9*9))
    zeros[pixel_list]=1
    print(zeros.reshape((9,9)))



from PySide2.QtWidgets import *
from PySide2 import QtGui
from CIDAN.fileHandling import createFileDialog


class Input(QFrame):
    def __init__(self, display_name, program_name, on_change_function, default_val,
                 tool_tip, display_tool_tip =False):
        super().__init__()
        self.program_name = program_name
        self.input_box_1 = None
        self.tool_tip = tool_tip
        self.default_val = default_val
        self.display_name = display_name
        self.layout_main = QVBoxLayout()
        self.layout_h = QHBoxLayout()
        temp_lable = QLabel()
        temp_lable.setText(display_name)
        temp_lable.setToolTip(tool_tip)
        self.layout_h.addWidget(temp_lable)
        self.on_change_function = on_change_function
        self.layout_main.addLayout(self.layout_h)
        if display_tool_tip:
            temp_lable_2 = QLabel()
            temp_lable_2.setText(tool_tip)
            self.layout_main.addWidget(temp_lable_2)

        self.setLayout(self.layout_main)

    def on_change(self, *args, **kwargs):
        try:
            self.on_change_function(self.program_name, self.current_state())
        except AssertionError as e:
            print(e)
            self.set_default_val()

    def current_state(self):
        pass
    def set_default_val(self):
        pass

class FloatInput(Input):
    def __init__(self, display_name, program_name, on_change_function, default_val,
                 tool_tip,
                 min, max, step, display_tool_tip =False
                 ):
        super().__init__(display_name, program_name, on_change_function, default_val,
                         tool_tip, display_tool_tip)

        self.input_box = QDoubleSpinBox()
        self.input_box.setMinimum(min)
        self.input_box.setMaximum(max)
        self.input_box.setMaximumWidth(50)
        self.input_box.setSingleStep(step)
        self.input_box.setValue(self.default_val)
        self.input_box.valueChanged.connect(self.on_change)
        self.input_box.setToolTip(self.tool_tip)
        self.layout_h.addWidget(self.input_box)

    def current_state(self):
        return self.input_box.value()
    def set_default_val(self):
        self.input_box.setValue(self.default_val)


class IntInput(Input):
    def __init__(self, display_name, program_name, on_change_function, default_val,
                 tool_tip,
                 min, max, step,  display_tool_tip =False
                 ):
        super().__init__(display_name, program_name, on_change_function, default_val,
                         tool_tip, display_tool_tip)

        self.input_box = QSpinBox()
        self.input_box.setMinimum(min)
        self.input_box.setMaximumWidth(50)
        self.input_box.setMaximum(max)
        self.input_box.setSingleStep(step)
        self.input_box.setValue(self.default_val)
        self.input_box.setToolTip(self.tool_tip)
        self.layout_h.addWidget(self.input_box)
        self.input_box.valueChanged.connect(self.on_change)

    def current_state(self):
        return self.input_box.value()
    def set_default_val(self):
        self.input_box.setValue(self.default_val)


# TODO implement way to check spatial boxes square
class BoolInput(Input):
    def __init__(self, display_name, program_name, on_change_function, default_val,
                 tool_tip, display_tool_tip =False):
        super().__init__(display_name, program_name, on_change_function, default_val,
                         tool_tip, display_tool_tip)

        self.input_box = QRadioButton()
        self.input_box.setMaximumWidth(50)
        self.input_box.setChecked(self.default_val)
        self.input_box.toggled.connect(self.on_change)
        self.input_box.setToolTip(self.tool_tip)
        self.layout_h.addWidget(self.input_box)

    def current_state(self):
        return self.input_box.isChecked()
    def set_default_val(self):
        self.input_box.setChecked(self.default_val)

class OptionInput(Input):
    def __init__(self, display_name, program_name, on_change_function, default_index,
                 tool_tip,
                 val_list, display_tool_tip =False):
        super().__init__(display_name, program_name, on_change_function,
                         default_index, tool_tip, display_tool_tip)

        self.input_box = QComboBox()
        self.val_list = val_list
        self.input_box.addItems(val_list)

        self.input_box.setCurrentIndex(self.default_val)
        self.input_box.setToolTip(self.tool_tip)
        self.input_box.currentIndexChanged.connect(self.on_change)
        self.layout_h.addWidget(self.input_box)

    def current_state(self):
        return self.val_list[self.input_box.currentIndex()]
    def set_default_val(self):
        self.input_box.setCurrentIndex(self.default_val)



class FileInput(Input):
    def __init__(self, display_name, program_name, on_change_function, default_val,
                 tool_tip, isFolder, forOpen):
        super().__init__(display_name, program_name, on_change_function,
                         default_val, tool_tip)
        self.isFolder = isFolder
        self.forOpen = forOpen
        self.path = ""
        self.current_location = QLabel()
        self.current_location.setText("")
        self.layout_h.addWidget(self.current_location)
        self.button = QPushButton()
        self.button.setText("Browse")
        self.button.clicked.connect(self.on_browse_button)
        self.button.setToolTip(tool_tip)
        self.layout_h.addWidget(self.button)

    def on_browse_button(self):
        self.path = createFileDialog(directory="~/Desktop", forOpen=self.forOpen,
                                     isFolder=self.isFolder)
        self.current_location.setText(self.path)

    def current_state(self):
        return self.path
class Int3DInput(Input):
    def __init__(self, display_name, program_name, on_change_function, default_val,
                 tool_tip,
                 min, max, step,  display_tool_tip =False
                 ):
        super().__init__(display_name, program_name, on_change_function, default_val,
                         tool_tip, display_tool_tip)

        self.input_box_1 = QSpinBox()
        self.input_box_1.setMinimum(min)
        self.input_box_1.setMaximumWidth(50)
        self.input_box_1.setMaximum(max)
        self.input_box_1.setSingleStep(step)
        self.input_box_1.setValue(self.default_val)
        self.input_box_1.setToolTip(self.tool_tip)
        self.layout_h.addWidget(self.input_box_1)
        self.input_box_1.valueChanged.connect(self.on_change)
        self.input_box_2 = QSpinBox()
        self.input_box_2.setMinimum(min)
        self.input_box_2.setMaximumWidth(50)
        self.input_box_2.setMaximum(max)
        self.input_box_2.setSingleStep(step)
        self.input_box_2.setValue(self.default_val)
        self.input_box_2.setToolTip(self.tool_tip)
        self.layout_h.addWidget(self.input_box_2)
        self.input_box_2.valueChanged.connect(self.on_change)
        self.input_box_3 = QSpinBox()
        self.input_box_3.setMinimum(min)
        self.input_box_3.setMaximumWidth(50)
        self.input_box_3.setMaximum(max)
        self.input_box_3.setSingleStep(step)
        self.input_box_3.setValue(self.default_val)
        self.input_box_3.setToolTip(self.tool_tip)
        self.layout_h.addWidget(self.input_box_3)
        self.input_box_3.valueChanged.connect(self.on_change)

    def current_state(self):
        return self.input_box_1.value()
    def set_default_val(self):
        self.input_box_1.setValue(self.default_val)



from CIDAN.LSSC.functions.embeddings import calc_D_sqrt, calc_D_neg_sqrt, calc_D_inv
import numpy as np
from dask import delayed
from scipy.sparse import linalg
import os
from PIL import Image
from CIDAN.LSSC.SpatialBox import combine_images
from CIDAN.LSSC.functions.pickle_funcs import pickle_save, pickle_exist, pickle_load
import logging
logger1 = logging.getLogger("CIDAN.LSSC.eigen")
@delayed
def gen_eigen_vectors(*, K: np.ndarray, num_eig: int, spatial_box_num:int, temporal_box_num: int)->np.ndarray:
    """Calculate Eigen Vectors given parts of the affinity matrix

    Parameters
    ----------

    K
        An affinity matrix K for the image
    num_eig
        number of eigen values to generate

    Returns
    -------
    A matrix of eigen vectors
    """
    D_inv, D_diag = calc_D_inv(K=K)
    P = D_inv*K
    D_neg_sqrt = calc_D_neg_sqrt(D_diag)
    P_transformed = calc_D_sqrt(D_diag)*P*D_neg_sqrt
    eig_values, eig_vectors_scaled = linalg.eigsh(
        P_transformed, num_eig, which="LM",
        return_eigenvectors=True)
    eig_vectors = np.flip(
        D_neg_sqrt * eig_vectors_scaled, axis=1)
    # print("Spatial Box {}, Time Step {}: Finished ".format(spatial_box_num, temporal_box_num))

    return eig_vectors
@delayed
def save_eigen_vectors(*, e_vectors, spatial_box_num: int, time_box_num:int, save_dir:
str):
    eigen_dir = os.path.join(save_dir, "eigen_vectors")
    if not os.path.isdir(eigen_dir):
        os.mkdir(eigen_dir)
    pickle_save(e_vectors, name="eigen_vectors_box_{}_{}.pickle".format(spatial_box_num,time_box_num),
                output_directory=eigen_dir, )
    return e_vectors
@delayed
def load_eigen_vectors(*, spatial_box_num: int, time_box_num: int, save_dir: str):
    eigen_dir = os.path.join(save_dir, "eigen_vectors")
    name = "eigen_vectors_box_{}_{}.pickle".format(spatial_box_num,time_box_num)
    logger1.debug("Eigen vector load: file name: {0}, directory: {1}".format(name,eigen_dir))
    vector = pickle_load(name=name,
                output_directory=eigen_dir)
    logger1.debug("Eigen spatial box {0}, time box {1} shape: {2}".format(spatial_box_num, time_box_num, vector.shape))
    return vector
@delayed
def save_embeding_norm_image(*, e_vectors, image_shape, save_dir, spatial_box_num):
    # print(save_dir)
    # embed_dir = os.path.join(save_dir, "embedding_norm_images")
    # e_vectors_squared = np.power(e_vectors, 2)
    # e_vectors_sum = np.sum(e_vectors_squared, axis=1)
    #
    # e_vectors_sum_rescaled = e_vectors_sum * (
    #         10.0 / e_vectors_sum.max())  # add histogram equalization
    #
    # img = Image.fromarray(
    #     np.reshape(e_vectors_sum_rescaled,
    #                image_shape[1:]) * 255).convert('L')
    # image_path = os.path.join(embed_dir, "embedding_norm_image_box_{}.png".format(
    #     spatial_box_num))
    # img.save(image_path)
    return e_vectors
def create_embeding_norm_multiple(*,  spatial_box_list, save_dir, num_time_steps):
    # print(save_dir)
    eigen_dir = os.path.join(save_dir, "eigen_vectors")
    e_vectors_list = []
    for spatial_box in spatial_box_list:
        temp = []
        for time_box_num in range(num_time_steps):
            temp.append(pickle_load("eigen_vectors_box_{}_{}.pickle".format(spatial_box.box_num, time_box_num), output_directory=eigen_dir))
        e_vectors_list.append(np.hstack(temp))
    embed_dir = os.path.join(save_dir, "embedding_norm_images")
    eigen_images = []
    for e_vectors, spatial_box in zip(e_vectors_list, spatial_box_list):
        e_vectors_squared = np.power(e_vectors, 2)
        e_vectors_sum = np.sum(e_vectors_squared, axis=1)
        e_vectors_sum_rescaled = e_vectors_sum * (
                9.0 / e_vectors_sum.max())  # add histogram equalization
        e_vectors_shaped = np.reshape(e_vectors_sum,
                   spatial_box.shape[1:])
        eigen_images.append(e_vectors_shaped)
    image = combine_images(spatial_box_list,eigen_images)



    img = Image.fromarray(image * (
                3.0 / image.max())*255).convert('L')
    image_path = os.path.join(embed_dir, "embedding_norm_image.png")
    img.save(image_path)
    return e_vectors_list


from PySide2.QtWidgets import *
from typing import Union, Any, List, Optional, cast, Tuple, Dict
from CIDAN.SettingsModule import *
from CIDAN.roiTools import *
from CIDAN.ImageViewModule import ImageViewModule
from CIDAN.Input import FileInput
from CIDAN.DataHandlerWrapper import *
from CIDAN.fileHandling import *
import pyqtgraph as pg
from CIDAN.ROIListModule import *
class Column(QWidget): # just used for stylesheets
    pass
class Tab(QWidget):
    def __init__(self, name, column_1: List[QFrame], column_2: List[QFrame], column_2_display=True):
        super().__init__()
        self.name = name
        self.column_1 = column_1
        self.column_2 = column_2
        self.setMinimumHeight(500)
        self.layout = QHBoxLayout() # Main layout class
        self.column_1_widget = Column()


        self.column_1_layout = QVBoxLayout()# Layout for column 1
        self.column_1_widget.setLayout(self.column_1_layout)
        self.column_1_widget.setStyleSheet("Column { border:1px solid rgb(50, 65, "
                                           "75);} ")
        for module in column_1:
            self.column_1_layout.addWidget(module)
        self.layout.addWidget(self.column_1_widget, stretch=1)
        if column_2_display:
            self.column_2_layout = QVBoxLayout() # Layout for column 2
            self.column_2_widget = Column()
            self.column_2_widget.setStyleSheet("Column { border:1px solid rgb(50, 65, "
                           "75);} ")
            self.column_2_widget.setLayout(self.column_2_layout)
            for module in column_2:
                self.column_2_layout.addWidget(module)
            self.layout.addWidget(self.column_2_widget, stretch=1)
        self.setLayout(self.layout)





class AnalysisTab(Tab):
    def __init__(self, main_widget):
        super().__init__("Analysis", column_1=[], column_2=[])



from PySide2.QtWidgets import *
from PySide2.QtGui import *
from PySide2.QtCore import *
import numpy as np
import sys

class MatrixSignal(QObject):
    def __init__(self):
        super().__init__()
    sig = Signal(np.matrix)
class StrSignal(QObject):
    def __init__(self):
        super().__init__()
    sig = Signal(str)

class BoolSignal(QObject):
    def __init__(self):
        super().__init__()

    sig = Signal(bool)
class Thread(QThread):
    def __init__(self, data_handler, parent=None):
        QThread.__init__(self, parent)
        self.exiting = False
        self.signal = MatrixSignal()
        self.data_handler = data_handler

    def run(self):
        pass

class PreprocessThread(Thread):
    def __init__(self, main_widget, button):
        super().__init__(main_widget.data_handler)
        self.main_widget = main_widget
        self.button = button
        self.signal.sig.connect(lambda x: self.endThread(x))

    def run(self):
        try:
            self.signal.sig.emit(self.data_handler.calculate_filters())
        except:
            print("Unexpected error:", sys.exc_info()[0])
            self.signal.sig.emit(np.matrix([0]))

    def runThread(self):

        if not any([x.isRunning() for x in self.main_widget.thread_list]):
            print("Starting preprocessing sequence")
            self.button.setEnabled(False)
            self.start()
    def endThread(self,image_data):
        self.button.setEnabled(True)
        if image_data.shape!=[1]:

            print("Finished preprocessing sequence")
            self.main_widget.preprocess_image_view.setImage(image_data)
class ROIExtractionThread(Thread):
    def __init__(self,main_widget, button, roi_list_module, roi_tab):
        super().__init__(main_widget.data_handler)
        self.signal = BoolSignal()
        self.roi_tab = roi_tab
        self.main_widget = main_widget
        self.roi_list_module = roi_list_module
        self.button = button
        self.signal.sig.connect(lambda x: self.endThread(x))
    def run(self):
        # TODO add  try except wrapper to this
        try:
            self.data_handler.calculate_roi_extraction()
            self.signal.sig.emit(True)
        except AssertionError as e:
            self.signal.sig.emit(False)


    def runThread(self):

        if not any([x.isRunning() for x in self.main_widget.thread_list]):
            print("Starting ROI extraction")
            self.button.setEnabled(False)
            self.start()
    def endThread(self, success):
        self.button.setEnabled(True)
        if success:
            print("Finished ROI extraction")
            self.roi_list_module.set_list_items(self.main_widget.data_handler.clusters)
            shape = self.main_widget.data_handler.edge_roi_image_flat.shape
            self.roi_tab.roi_image_flat = np.hstack([self.main_widget.data_handler.edge_roi_image_flat,
                                                np.zeros(shape),
                                                np.zeros(shape)])
            self.roi_tab.select_image_flat = np.zeros([shape[0],3])

            self.roi_tab.updateImageDisplay(new= True)




import numpy as np
from dask import array
def calculate_temporal_correlation(dataset):
    shape = dataset.shape
    default_val = 0 #np.average(dataset)
    dataset_flat = array.from_array(dataset.transpose([1,2,0]).reshape([shape[1]*shape[2],-1]))

    corr_up = test_func(default_val, dataset_flat,shape,dataset).compute()
    shift_down = np.hstack([dataset[:,1:,:],np.full((shape[0],1, shape[2]), default_val)]).transpose([1,2,0]).reshape([shape[1]*shape[2],-1])
    corr_down = np.average(dataset_flat * shift_down, axis=1)
    del shift_down
    shift_right = np.dstack([np.full((shape[0],shape[1], 1),default_val),dataset[:,:,:-1]]).transpose([1,2,0]).reshape([shape[1]*shape[2],-1])
    corr_right = np.average(dataset_flat*shift_right,axis=1)
    del shift_right
    shift_left = np.dstack([dataset[:,:,:-1],np.full((shape[0],shape[1], 1),default_val)]).transpose([1,2,0]).reshape([shape[1]*shape[2],-1])
    corr_left = np.average(dataset_flat*shift_left,axis=1)
    del shift_left
    avg_correlation = np.average(np.vstack([corr_up,corr_down,corr_left,corr_right]),axis=0)
    avg_correlation_image = avg_correlation.reshape([shape[1],shape[2]])
    return avg_correlation_image
def test_func(default_val, dataset_flat,shape,dataset):
    shift_up = array.hstack([array.zeros((shape[0], 1, shape[2])), dataset[:, :-1, :]]).transpose(
        [1, 2, 0]).reshape([shape[1] * shape[2], -1])

    shift_up_mult = dataset_flat * shift_up
    del shift_up
    return array.mean(shift_up_mult, axis=1)


from PySide2 import QtCore

from CIDAN.Tab import Tab
from PySide2.QtWidgets import *
from PySide2.QtCore import *
import numpy as np
import pyqtgraph as pg
from CIDAN.DataHandlerWrapper import ROIExtractionThread
from CIDAN.SettingsModule import roi_extraction_settings
from CIDAN.ROIListModule import ROIListModule
from CIDAN.Input import OptionInput
from CIDAN.LSSC.functions.roi_extraction import combine_rois
class ROIExtractionTab(Tab):
    """Class controlling the ROI Extraction tab, inherits from Tab


        Attributes
        ----------
        main_widget : MainWidget
            A reference to the main widget
        data_handler : DataHandler
            A reference to the main DataHandler of MainWidget
        click_event : bool
            A bool that keeps track of whether a click event is currently happening used
            by roi_click_event and select_roi
        time_plot : pg.PlotWidget
            the plot for the time traces
        roi_list_module : ROIListModule
            The module the controlls the list of ROIs
        thread : ROIExtractionThread
            The thread that runs the roi extraction process
        foreground_slider : QSlider
            slider that determines intensity of background image+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

        """
    def __init__(self,main_widget):

        self.main_widget = main_widget
        self.data_handler = main_widget.data_handler
        self.click_event = False
        self.add_image = False
        self.outlines = True

        # part about selecting pixels

        self.select_pixel_on = False  # whether you can currently select pixels in the image
        self.brush_size = 0
        self.current_selected_pixels_list = []  # list of currently selected pixels in their 1d number format
        self.current_selected_pixels_mask = np.zeros((self.data_handler.shape[1],
                                                      self.data_handler.shape[
                                                          2]), dtype=bool)  # mask 1 when selected 0 when not
        self.previous_values = {}
        self.select_pixel_color = [0, 255, 0]
        self.select_mode = "add"  # possibilities add and subtract from current selection
        # overload the image click and drag properites on the image item
        self.image_item = self.main_widget.roi_image_view.image_view.getImageItem()
        self.image_item.mouseClickEvent = lambda x: self.roi_view_click(x)
        self.image_item.mouseDragEvent = lambda x: self.roi_view_drag(x)
        # This part creates the top left settings/roi list view in two tabs
        tab_selector_roi = QTabWidget()
        tab_selector_roi.setStyleSheet("QTabWidget {font-size: 20px;}")
        # This is the second tab
        roi_modification_tab = QWidget()
        roi_modification_tab.setStyleSheet("margin:0px; padding: 0px;")

        roi_modification_tab_layout = QVBoxLayout()
        roi_modification_tab.setLayout(roi_modification_tab_layout)
        self.roi_list_module = ROIListModule(main_widget.data_handler, self)
        roi_modification_tab_layout.addWidget(self.roi_list_module)
        roi_modification_button_top_layout = QHBoxLayout()
        roi_modification_tab_layout.addLayout(roi_modification_button_top_layout)
        clear_from_selection = QPushButton(text="Clear Selection")
        clear_from_selection.clicked.connect(lambda x: self.clearPixelSelection())
        add_to_roi = QPushButton(text="New ROI from Selection")



        roi_modification_button_top_layout.addWidget(clear_from_selection)
        roi_modification_button_top_layout.addWidget(add_to_roi)
        # roi_modification_button_top_layout.addWidget(subtract_from_roi)
        # painter_options = OptionInput("Pixel Selector Brush:", "",
        #                               lambda x,y: self.setSelectorBrushType(y),
        #                               default_index=0, tool_tip="",
        #                               val_list=["Off", "Add to Selection",
        #                                         "Subtract from Selection"])
        painter_button_group = QButtonGroup()
        off_button = QRadioButton(text="Off")
        off_button.setChecked(True)
        # off_button.connect(lambda x: self.setSelectorBrushType("off"))
        on_button = QRadioButton(text="Add to Selection")
        # on_button.connect(lambda x: self.setSelectorBrushType("Add to Selection"))
        sub_button = QRadioButton(text="Subtract from Selection")
        # sub_button.connect(lambda x: self.setSelectorBrushType("off"))
        painter_button_group.addButton(off_button)
        painter_button_group.addButton(on_button)
        painter_button_group.addButton(sub_button)
        off_button.clicked.connect(
            lambda x: self.setSelectorBrushType("off"))
        on_button.clicked.connect(
            lambda x: self.setSelectorBrushType("add"))
        sub_button.clicked.connect(
            lambda x: self.setSelectorBrushType("subtract"))

        painter_layout = QHBoxLayout()
        painter_layout.addWidget(QLabel(text="Selector Brush: "))
        painter_layout.addWidget(off_button)
        painter_layout.addWidget(on_button)
        painter_layout.addWidget(sub_button)
        roi_modification_tab_layout.addLayout(painter_layout)
        brush_size_options = OptionInput("Brush Size:", "",
                                         lambda x, y: self.setBrushSize(y), 0,
                                         "Sets the brush size", ["1","3","5","7","9",
                                                                 "11","15","21","27",
                                                                 "35"])
        roi_modification_tab_layout.addWidget(brush_size_options)






        # Start of first tab with the process button
        process_button = QPushButton()
        process_button.setText("Apply Settings")
        self.thread = ROIExtractionThread(main_widget, process_button, self.roi_list_module,self)
        self.main_widget.thread_list.append(self.thread)
        process_button.clicked.connect(lambda: self.thread.runThread())
        self.roi_settings = QWidget()
        self.roi_settings_layout = QVBoxLayout()
        self.roi_settings.setLayout(self.roi_settings_layout)
        self.roi_settings_layout.addWidget(roi_extraction_settings(main_widget))
        self.roi_settings_layout.addWidget(process_button)
        # adding the tabs to the
        tab_selector_roi.addTab(self.roi_settings, "ROI Creation")
        tab_selector_roi.addTab(roi_modification_tab, "ROI Modification")

        # Initialization of the background and rois
        self.current_foreground_intensity = 1
        self.set_background("", "Max Image", update_image=False)
        if self.main_widget.data_handler.rois_loaded:
            self.thread.endThread(True)



        # Image and time trace settings window


        display_settings_layout = QVBoxLayout()

        display_settings = QWidget()
        display_settings.setLayout(display_settings_layout)
        image_chooser = OptionInput("ROI Display type::", "",
                                         on_change_function=self.set_image,
                                         default_index=0,
                                         tool_tip="Choose background to display",
                                         val_list=["Outlines", "Blob"])

        display_settings_layout.addWidget(image_chooser)






        background_chooser = OptionInput("Background:", "",
                                              on_change_function=self.set_background, default_index=2,
                                              tool_tip="Choose background to display",
                                              val_list=["Blank Image", "Mean Image", "Max Image", "Temporal Correlation Image", "Eigen Norm Image"])

        display_settings_layout.addWidget(background_chooser)
        background_slider_layout = QHBoxLayout()
        background_slider_layout.addWidget(QLabel("0"))
        self.foreground_slider = QSlider(Qt.Horizontal)
        self.foreground_slider.setMinimum(0)

        self.foreground_slider.setMaximum(100)
        self.foreground_slider.setSingleStep(1)
        self.foreground_slider.valueChanged.connect(self.intensity_slider_changed)
        self.foreground_slider.setValue(80)
        background_slider_layout.addWidget(self.foreground_slider)
        background_slider_layout.addWidget(QLabel("10"))

        display_settings_layout.addWidget(QLabel("Change foreground intensity:"))
        display_settings_layout.addLayout(background_slider_layout)

        # Initialize the tab with each column

        # this is to override how we do the column 2 to replace it with a split view
        tab_selector_time_trace = QTabWidget()
        tab_selector_time_trace.setStyleSheet("QTabWidget {font-size: 20px;}")
        tab_selector_time_trace.setMaximumHeight(200)
        self.time_plot = pg.PlotWidget()
        self.time_plot.showGrid(x=True, y=True, alpha=0.3)
        tab_selector_time_trace.addTab(self.time_plot, "Time Trace Plot")
        tab_selector_time_trace.addTab(QWidget(), "Time Trace Settings")



        roi_view_tabs = QTabWidget()
        roi_view_tabs.setStyleSheet("QTabWidget {font-size: 20px;}")
        # roi_view_tabs.setStyleSheet(
        #     "QButton, QLabel, QSlider {padding: 5px; margin: 5px;}")
        # roi_view_tabs.setStyleSheet("QTabWidget {font-size: 20px;}")
        self.main_widget.roi_image_view.setStyleSheet("margin:0px; border:0px  solid rgb(50, 65, "
                           "75); padding: 0px;")
        roi_view_tabs.addTab(self.main_widget.roi_image_view, "ROI Display")
        roi_view_tabs.addTab(display_settings, "Display Settings")
        self.column_2 = [roi_view_tabs, tab_selector_time_trace ]
        super().__init__("ROI Extraction",
                         column_1=[tab_selector_roi],
                         column_2=self.column_2, column_2_display=True)
        # self.setStyleSheet("SettingsModule { border:1px solid rgb(50, 65, "
        #                    "75);} ")
        # column_2_split = QSplitter(Qt.Vertical)  # Layout for column 2 with split part
        # for module in self.column_2:
        #     column_2_split.addWidget(module)
        # column_2_split.setSizes([400, 100])
        #
        #
        # column_2_layout_box.addWidget(column_2_split)
        # # column_2_layout_box.addWidget(tab_selector_image)
        # self.layout.addLayout(column_2_layout_box)

    def setSelectorBrushType(self,type):

        if type == "off":
            self.select_pixel_on = False
        else:
            self.select_pixel_on = True
            self.select_mode = type


    def modify_roi(self,  roi_num,add_subtract="add"):
        """
        Add/subtracts the currently selected pixels from an ROI
        Parameters
        ----------
        roi_num roi to modify starting at 1
        add_subtract either add or subtract depending on operation wanted

        Returns
        -------
        Nothing
        """
        shape = self.main_widget.data_handler.edge_roi_image_flat.shape
        roi_num = roi_num-1
        if add_subtract == "add":
            self.data_handler.clusters[roi_num] = combine_rois(self.data_handler.clusters[roi_num], self.current_selected_pixels_list)
            self.data_handler.gen_roi_display_variables()
            self.data_handler.calculate_time_trace(roi_num)
        if add_subtract == "subtract":
            self.data_handler.clusters[roi_num] = [x for x in self.data_handler.clusters[roi_num] if x not in self.current_selected_pixels_list]
            self.data_handler.gen_roi_display_variables()
            self.data_handler.calculate_time_trace(roi_num)
        if self.outlines:
            self.roi_image_flat = np.hstack([self.data_handler.edge_roi_image_flat,
                                             np.zeros(shape),
                                             np.zeros(shape)])


        else:
            self.roi_image_flat = self.main_widget.data_handler.pixel_with_rois_color_flat


        self.select_image_flat = np.zeros([shape[0], 3])
        self.clearPixelSelection(update_display=False)
        self.updateImageDisplay()
    def draw(self,pos):
        pass
    def intensity_slider_changed(self):
        self.current_foreground_intensity= 10-(float(self.foreground_slider.value()) / 10)
        self.updateImageDisplay()
    def set_background(self, name,func_name, update_image=True):
        # Background refers to the image behind the rois
        shape = self.main_widget.data_handler.shape
        if func_name == "Mean Image":
            self.current_background = self.main_widget.data_handler.mean_image.reshape([-1, 1])
        if func_name == "Max Image":
            self.current_background = self.main_widget.data_handler.max_image.reshape([-1, 1])
        if func_name == "Blank Image":
            self.current_background = np.zeros([shape[1]*shape[2],1])
        if func_name == "Temporal Correlation Image":
            self.current_background = self.data_handler.temporal_correlation_image.reshape([-1, 1])
        if func_name == "Eigen Norm Image":
            self.current_background = self.data_handler.eigen_norm_image.reshape([-1, 1])

        if update_image:
            self.updateImageDisplay()

    def set_image(self, name, func_name, update_image=True):
        # Background refers to the image behind the rois
        shape = self.main_widget.data_handler.edge_roi_image_flat.shape
        if func_name == "Outlines":
            self.outlines=True
            self.roi_image_flat = np.hstack([self.data_handler.edge_roi_image_flat,
                                             np.zeros(shape),
                                             np.zeros(shape)])
        if func_name == "Blob":
            self.outlines = False
            self.roi_image_flat = self.main_widget.data_handler.pixel_with_rois_color_flat

        if update_image:
            self.updateImageDisplay()
    def updateImageDisplay(self, new=False):
        # new is to determine whether the zoom should be saved
        # TODO add in update with image paint layer
        shape = self.main_widget.data_handler.dataset_filtered.shape
        if not hasattr(self, "select_image_flat"):
            self.select_image_flat = np.zeros([shape[1]*shape[2], 3])
        range_list=self.main_widget.roi_image_view.image_view.view.viewRange()
        print(range_list)
        shape = self.main_widget.data_handler.dataset_filtered.shape
        background_max = self.current_background.max()
        background_image_scaled = (self.current_foreground_intensity * 255 / (background_max if background_max != 0 else 1)) * self.current_background
        background_image_scaled_3_channel = np.hstack([background_image_scaled,background_image_scaled ,background_image_scaled])



        if new:
            # if self.add_image:
            combined = self.roi_image_flat + background_image_scaled_3_channel + self.select_image_flat

            # else:
            #     combined = background_image_scaled + self.select_image_flat
            #     mask = np.any(self.roi_image_flat != [0, 0, 0], axis=1)
            #     combined[mask] = self.roi_image_flat[mask]
            combined_reshaped = combined.reshape((shape[1], shape[2], 3))
            self.main_widget.roi_image_view.setImage(combined_reshaped)
            self.clearPixelSelection(update_display= False)
        else:
            self.image_item.image = background_image_scaled_3_channel.reshape((shape[1], shape[2], 3))
            self.image_item.updateImage(autoLevels=True)

            # if self.add_image:
            combined = (self.roi_image_flat + self.select_image_flat).reshape(
                (shape[1], shape[2], 3))
            self.image_item.image += combined
            self.image_item.image[self.current_selected_pixels_mask] += self.select_pixel_color

            # else:
            #     combined = self.select_image_flat+self.roi_image_flat
            #     combined_reshaped = combined.reshape((shape[1], shape[2], 3))
            #     mask = np.any(combined != [0, 0, 0], axis=1).reshape((shape[1], shape[2]))
            #
            #     self.image_item.image[mask] = combined_reshaped[mask]
            self.image_item.updateImage(autoLevels=False)

            # self.main_widget.roi_image_view.image_view.view.setRange(xRange=range_list[0],
            #                                                      yRange=range_list[1])
            # range_list = self.main_widget.roi_image_view.image_view.view.viewRange()
            # print(range_list)

        pass

    def selectRoi(self, num):

        color_select = (245, 249, 22)
        color_roi = self.main_widget.data_handler.color_list[(num-1) % len(self.main_widget.data_handler.color_list)]
        shape = self.main_widget.data_handler.dataset_filtered.shape
        self.select_image_flat[self.main_widget.data_handler.clusters[num - 1]] = color_select
        self.updateImageDisplay()



        pen = pg.mkPen(color=color_roi, width=3)
        self.time_plot.plot(self.main_widget.data_handler.get_time_trace(num), pen=pen)
        self.time_plot.enableAutoRange(axis=0)
    def deselectRoi(self, num):

        color = self.main_widget.data_handler.color_list[(num-1) % len(self.main_widget.data_handler.color_list)]
        shape = self.main_widget.data_handler.dataset_filtered.shape
        shape_flat = self.data_handler.edge_roi_image_flat.shape
        self.select_image_flat[self.main_widget.data_handler.clusters[num - 1]] = color if not self.outlines \
            else np.hstack([self.data_handler.edge_roi_image_flat,
                                                np.zeros(shape_flat),
                                                np.zeros(shape_flat)])[self.main_widget.data_handler.clusters[num - 1]]
        self.updateImageDisplay()
        self.time_plot.clear()
        self.time_plot.enableAutoRange(axis=0)
        for num2, x in zip(range(1,len(self.roi_list_module.roi_check_list)),self.roi_list_module.roi_check_list):
            if x:
                color_roi = self.main_widget.data_handler.color_list[
                    (num2 - 1) % len(self.main_widget.data_handler.color_list)]

                pen = pg.mkPen(color=color_roi, width=3)
                self.time_plot.plot(
                    self.main_widget.data_handler.get_time_trace(num2), pen=pen)

    def zoomRoi(self, num):
        """
        Zooms in to a certain roi
        Parameters
        ----------
        num : int
            roi num starts at 1

        Returns
        -------
        Nothing
        """
        num = num-1

        max_cord = self.main_widget.data_handler.cluster_max_cord_list[num] + 15

        min_cord = self.main_widget.data_handler.cluster_min_cord_list[num] - 15

        self.main_widget.roi_image_view.image_view.getView().setXRange(min_cord[1],
                                                                       max_cord[1])
        self.main_widget.roi_image_view.image_view.getView().setYRange(min_cord[0],
                                                                       max_cord[0])
    def roi_view_click(self, event):
        if event.button() == QtCore.Qt.RightButton:
            if self.image_item.raiseContextMenu(event):
                event.accept()
        event.accept()
        pos = event.pos()

        x = int(pos.x())
        y = int(pos.y())
        if self.select_pixel_on:
            self.pixel_paint(x,y)
        else:
            self.click_event = True
            pixel_with_rois_flat = self.main_widget.data_handler.pixel_with_rois_flat
            shape = self.main_widget.data_handler.dataset_filtered.shape
            roi_num = int(pixel_with_rois_flat[shape[2] * x + y])
            # TODO change to int
            if roi_num != 0:
                self.roi_list_module.set_current_select(roi_num)


    def roi_view_drag(self,event):
        # if event.button() == QtCore.Qt.RightButton:
        #     if self.image_item.raiseContextMenu(event):
        #         event.accept()
        event.accept()
        pos = event.pos()

        x = int(pos.x())
        y = int(pos.y())
        if self.select_pixel_on:
            self.pixel_paint(x,y)


    def pixel_paint(self,x,y):
        try:
            if self.select_mode == "add":
                shape = self.main_widget.data_handler.dataset_filtered.shape

                for x_dif in range(self.brush_size*2+1):
                    for y_dif in range(self.brush_size*2+1):
                        x_new = x-self.brush_size-1+x_dif
                        y_new = y-self.brush_size-1+y_dif
                        if shape[2] * x_new + y_new not in self.current_selected_pixels_list:
                            self.image_item.image[x_new, y_new] += [0, 255, 0]
                            self.current_selected_pixels_list.append(shape[2] * x_new + y_new)
                            self.current_selected_pixels_mask[x_new, y_new] = True


            if self.select_mode == "subtract":
                shape = self.main_widget.data_handler.dataset_filtered.shape
                for x_dif in range(self.brush_size * 2 + 1):
                    for y_dif in range(self.brush_size * 2 + 1):
                        x_new = x - self.brush_size - 1 + x_dif
                        y_new = y - self.brush_size - 1 + y_dif
                        if  shape[2] * x_new + y_new in  self.current_selected_pixels_list:
                            self.image_item.image[x_new, y_new] -= [0, 255, 0]
                            self.current_selected_pixels_list.remove(shape[2] * x_new + y_new)
                            self.current_selected_pixels_mask[x_new,y_new] = False
            self.image_item.updateImage()
        except IndexError:
            pass

        pass # TODO use slicing to update pixel based on current thing
    def clearPixelSelection(self,update_display= True):
        shape = self.main_widget.data_handler.dataset_filtered.shape
        self.current_selected_pixels_mask = np.zeros([shape[1],shape[2]],dtype=bool)
        self.current_selected_pixels_list = []
        if update_display:
            self.updateImageDisplay()
    def check_pos_in_image(self, x,y):
        pass
        # TODO add in way to check if in image
    def setBrushSize(self, size):
        """
        Sets the brush size

        self.brush_size is the additional size on all dimensions in addition to middle
        point
        Parameters
        ----------
        size from option input

        Returns
        -------
        nothing
        """
        self.brush_size= int((int(size)-1)/2)


from PySide2.QtWidgets import *
import sys
import qdarkstyle
from CIDAN.SettingBlockModule import *

class SettingsModule(QFrame):
    def __init__(self, name, *args, show_name=True):
        super().__init__()


        self.setting_block_list = args
        self.layout = QVBoxLayout()
        if show_name:
            self.setStyleSheet("SettingsModule { border:1px solid rgb(50, 65, "
                               "75);} ")
            self.header = QLabel()
            self.header.setText(name)
            self.header.setStyleSheet("font-size: 20px")
            self.layout.addWidget(self.header)

        self.setting_block_layout = QToolBox()
        for block in self.setting_block_list:
            self.setting_block_layout.addItem(block, block.name)
        self.layout.addWidget(self.setting_block_layout)
        self.setLayout(self.layout)
def preprocessing_settings(main_widget):
    return SettingsModule( "Preprocessing Settings", dataset_setting_block(main_widget),
                          filter_setting_block(main_widget))
def roi_extraction_settings(main_widget):
    return SettingsModule( "ROI Extraction Settings", multiprocessing_settings_block(main_widget), roi_extraction_settings_block(main_widget), roi_advanced_settings_block(main_widget), show_name=False)
if __name__ == "__main__":
    app = QApplication([])

    widget = preprocessing_settings()
    widget.setStyleSheet(qdarkstyle.load_stylesheet())
    widget.show()


    sys.exit(app.exec_())


import os

os.environ["OMP_NUM_THREADS"] = "10"  # export OMP_NUM_THREADS=4
os.environ["OPENBLAS_NUM_THREADS"] = "10"  # export OPENBLAS_NUM_THREADS=4
os.environ["MKL_NUM_THREADS"] = "10"  # export MKL_NUM_THREADS=6
os.environ["VECLIB_MAXIMUM_THREADS"] = "10"  # export VECLIB_MAXIMUM_THREADS=4
os.environ["NUMEXPR_NUM_THREADS"] = "10"  # export NUMEXPR_NUM_THREADS=6
import scipy.sparse as sparse
import hnswlib as hnsw
import numpy as np

from dask import delayed
@delayed
def calc_affinity_matrix(*, pixel_list: np.matrix, metric: str, knn:int,
                         accuracy:int, connections: int, normalize_w_k: int,
                         num_threads: 
        int, spatial_box_num:int, temporal_box_num: int):
    """
    Calculates an pairwise affinity matrix for the image stack
    Parameters
    ----------
    pixel_list: a 2d np array of pixels with dim 0 being a list of pixels and dim 1 being pixel values over time
    metric: can be "l2" squared l2, "ip" Inner product, "cosine" Cosine similarity
    knn: number of nearest neighbors to search for
    accuracy: time of construction vs acuracy tradeoff
    connections: max number of outgoing connections
    num_threads: number of threads to use
    normalize_w_k: kth clostest neighbor for autotune

    Returns
    -------
    affinity matrix
    """
    # TODO make connections value scale based on available memory

    dim = pixel_list.shape[1]
    num_elements = pixel_list.shape[0]
    print("Spatial Box {}, Time Step {}: Started Processing".format(spatial_box_num, temporal_box_num))
    p = hnsw.Index(space=metric, dim=dim)
    p.init_index(max_elements=num_elements, ef_construction=accuracy,
                 M=connections)
    p.add_items(pixel_list, num_threads=num_threads)
    indices, distances = p.knn_query(pixel_list, k=knn,
                                     num_threads=num_threads)  # lazy random walk means it returns distance of zero for same point

    reformat_indicies_x = np.repeat(np.arange(0, num_elements, 1), knn)
    reformat_indicies_y = np.reshape(indices, (-1))
    reformat_distances = np.reshape(distances, (-1)) # this might have broke stuff idk

    scale_factor_indices = np.repeat(distances[:,normalize_w_k], knn)
    scale_factor_2_per_distances = scale_factor_indices[reformat_indicies_x] * scale_factor_indices[
        reformat_indicies_y]
    reformat_distances_scaled = np.exp(
        -reformat_distances / scale_factor_2_per_distances)

    return sparse.csr_matrix(sparse.coo_matrix(
        (
            reformat_distances_scaled,
            (reformat_indicies_x, reformat_indicies_y)),
        shape=(num_elements, num_elements)))


def calc_D_inv(K):
    """Calculates a scaling diagonal matrix D to rescale eigen vectors

    Parameters
    ----------
    K the sparse matrix K for the pairwise affinity

    Returns
    -------
    a sparse matrix with type csr, and D's diagonal values
    """
    dim = K.shape[0]
    # D_diag = np.nan_to_num(1/K.sum(axis=1), nan=0.0, posinf =0, neginf=0) #add small epsilon to each row in K.sum()
    D_diag = 1 / K.sum(axis=1)

    D_sparse = sparse.dia_matrix((np.reshape(D_diag, [1, -1]), [0]),
                                 (dim, dim))
    return sparse.csr_matrix(D_sparse), D_diag


def calc_laplacian(P_sparse, dim):
    I_sparse = sparse.identity(dim, format="csr")
    laplacian_sparse = I_sparse - P_sparse
    return laplacian_sparse


def calc_D_sqrt(D_diag):
    dim = D_diag.shape[0]
    D_sqrt = sparse.csr_matrix(
        sparse.dia_matrix((np.reshape(np.power(D_diag, .5), [1, -1]), [0]),
                          (dim, dim)))
    return D_sqrt


def calc_D_neg_sqrt(D_diag):
    dim = D_diag.shape[0]
    D_neg_sqrt = sparse.csr_matrix(
        sparse.dia_matrix(
            (np.reshape(np.power(D_diag, -.5), [1, -1]), [0]),
            (dim, dim)))
    return D_neg_sqrt

def embed_eigen_norm(eigen_vectors): # embedding norm not just embeding
    pixel_embedings = np.sum(np.power(eigen_vectors, 2),
                             axis=1)
    return pixel_embedings



from setuptools import setup, find_packages

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
    name='CIDAN',
    version='0.1.5',
    description='CIDAN-Calcium Imaging Data ANalysis',
    license="MIT",
    long_description=long_description,
    author='Sam Schickler',
    author_email='sschickl@ucsd.edu',
    # url="http://www.foopackage.com/",
    packages=find_packages(),
    install_requires=["QtPy","QDarkStyle", "Pyside2","numpy", "dask[complete]", "matplotlib","scipy","tiffile","scikit-image","hnswlib", "pillow"],  # TODO add external packages as dependencies
    scripts=[
        "CIDAN/LSSC/process_data.py",
        "CIDAN/LSSC/SpatialBox.py",
        'CIDAN/LSSC/functions/data_manipulation.py',
        "CIDAN/LSSC/functions/eigen.py",
        "CIDAN/LSSC/functions/embeddings.py",
        "CIDAN/LSSC/functions/pickle_funcs.py",
        "CIDAN/LSSC/functions/roi_extraction.py",
        "CIDAN/LSSC/functions/save_test_images.py",
        "CIDAN/LSSC/functions/temporal_correlation.py",
        "CIDAN/ConsoleWidget.py",
        "CIDAN/DataHandler.py",
        'CIDAN/DataHandlerWrapper.py',
        'CIDAN/fileHandling.py',
        'CIDAN/FileOpenTab.py',
        'CIDAN/ImageViewModule.py',
        'CIDAN/Input.py',
        'CIDAN/MainWindow.py',
        'CIDAN/PreprocessingTab.py',
        'CIDAN/ROIExtractionTab.py',
        'CIDAN/ROIItemModule.py',
        'CIDAN/ROIListModule.py',
        'CIDAN/roiTools.py',
        'CIDAN/SettingBlockModule.py',
        'CIDAN/SettingsModule.py',
        'CIDAN/Tab.py'
    ]
)



import json
import os
from typing import Dict
from PIL import Image
from skimage import feature
import numpy as np
from CIDAN.LSSC.functions.data_manipulation import load_filter_tif_stack, filter_stack, \
    reshape_to_2d_over_time, pixel_num_to_2d_cord
from CIDAN.LSSC.functions.temporal_correlation import calculate_temporal_correlation
from CIDAN.LSSC.functions.pickle_funcs import *
from CIDAN.LSSC.process_data import process_data
import logging
logger1 = logging.getLogger("CIDAN.DataHandler")
class DataHandler:
    global_params_default = {
        "save_intermediate_steps": True,
        "need_recalc_dataset_params": True,
        "need_recalc_filter_params": True,
        "need_recalc_box_params": True,
        "need_recalc_eigen_params": True,
        "need_recalc_roi_extraction_params": True,
        "num_threads": 1
    }

    dataset_params_default = {
        "dataset_path": "",
        "slice_stack": False,
        "slice_every": 3,
        "slice_start": 0
    }

    filter_params_default = {
        "median_filter": False,
        "median_filter_size": 3,
        "z_score": False

    }
    box_params_default = {
        "total_num_time_steps": 1,
        "total_num_spatial_boxes": 4,
        "spatial_overlap": 15
    }
    eigen_params_default = {
        "eigen_vectors_already_generated": False,
        "num_eig": 24,
        "normalize_w_k": 2,
        "metric": "l2",
        "knn": 20,
        "accuracy": 39,
        "connections": 40,

    }
    roi_extraction_params_default = {
        "elbow_threshold_method": True,
        "elbow_threshold_value": 1,
        "eigen_threshold_method": True,
        "eigen_threshold_value": .5,
        "num_eigen_vector_select": 5,
        "merge_temporal_coef": .05,
        "roi_size_min": 30,
        "roi_size_max": 600,
        "merge": True,
        "num_rois": 25,
        "fill_holes": True,
        "refinement": True,
        "max_iter": 1000,
    }
    def __init__(self, data_path, save_dir_path, save_dir_already_created):
        # TODO Create two initialization methods one with default parameters and one
        #  with loading save dir parameters
        # TODO Check if parrelization is working
        self.color_list = [(218, 67, 34),
                           (132, 249, 22), (22, 249, 140), (22, 245, 249),
                           (22, 132, 249), (224, 22, 249), (249, 22, 160)]

        self.save_dir_path = save_dir_path
        self.rois_loaded = False
        if save_dir_already_created:
            valid = self.load_param_json()
            self.calculate_filters()
            #

            if self.rois_exist:
                self.load_rois()
                self.calculate_time_traces()
                self.rois_loaded = True
            if not valid:
                raise FileNotFoundError("Save directory not valid")
        else:
            self.global_params = DataHandler.global_params_default.copy()

            self.dataset_params = DataHandler.dataset_params_default.copy()
            self.dataset_params["dataset_path"] = data_path

            self.filter_params = DataHandler.filter_params_default.copy()
            self.box_params = DataHandler.box_params_default.copy()
            self.eigen_params = DataHandler.eigen_params_default.copy()
            self.roi_extraction_params = DataHandler.roi_extraction_params_default.copy()
            valid = self.create_new_save_dir()
            if not valid:
                raise FileNotFoundError("Please chose an empty directory for your " +
                                        "save directory")
            self.time_traces = []

    def __del__(self):
        for x in self.__dict__.items():
            self.__dict__[x] = None
    @property
    def param_path(self):
        return os.path.join(self.save_dir_path, "parameters.json")
    @property
    def eigen_vectors_exist(self):
        eigen_dir = os.path.join(self.save_dir_path, "eigen_vectors")

        file_names = ["eigen_vectors_box_{}_{}.pickle".format(spatial_box_num,time_box_num)
                      for spatial_box_num in range(self.box_params["total_num_spatial_boxes"])
                      for time_box_num in range(self.box_params["total_num_time_steps"])]
        return all(pickle_exist(x,output_directory=eigen_dir) for x in file_names)
    @property
    def rois_exist(self):
        return pickle_exist("rois",output_directory=self.save_dir_path)
    def load_rois(self):
        if pickle_exist("rois",output_directory=self.save_dir_path):
            self.clusters = pickle_load("rois", output_directory=self.save_dir_path)
            self.gen_roi_display_variables()
    def save_rois(self, rois):
        if os.path.isdir(self.save_dir_path):
            pickle_save(rois, "rois",output_directory=self.save_dir_path)
    def load_param_json(self):

        try:
            with open(self.param_path, "r") as f:
                all_params = json.loads(f.read())
            self.global_params = all_params["global_params"]
            self.dataset_params = all_params["dataset_params"]
            self.filter_params = all_params["filter_params"]
            self.box_params = all_params["box_params"]
            self.eigen_params = all_params["eigen_params"]
            self.roi_extraction_params = all_params["roi_extraction_params"]
            return True
        except KeyError:
            raise KeyError("Please Choose a valid parameter file")
        except FileNotFoundError:
            raise FileNotFoundError("Can't find parameter file")
        except NameError:
            raise FileNotFoundError("Can't find parameter file")

    def save_new_param_json(self):
        try:
            with open(self.param_path, "w") as f:
                all_params = {
                    "global_params": self.global_params,
                    "dataset_params": self.dataset_params,
                    "filter_params": self.filter_params,
                    "box_params": self.box_params,
                    "eigen_params": self.eigen_params,
                    "roi_extraction_params": self.roi_extraction_params
                }
                f.truncate(0)
                f.write(json.dumps(all_params))
        except:
            raise FileNotFoundError("Error saving parameters, please restart software")
        pass

    def create_new_save_dir(self):
        try:
            if not os.path.isdir(self.save_dir_path):

                os.mkdir(self.save_dir_path)
            eigen_vectors_folder_path = os.path.join(self.save_dir_path,
                                                     "eigen_vectors/")
            if not os.path.isdir(eigen_vectors_folder_path):
                os.mkdir(eigen_vectors_folder_path)
            embedding_images_path = os.path.join(self.save_dir_path,
                                                 "embedding_norm_images/")
            if not os.path.isdir(embedding_images_path):
                os.mkdir(embedding_images_path)
            return True
        except:
            raise FileNotFoundError("Couldn't create folder please try again")

    def change_global_param(self, param_name, new_value):
        if param_name in self.global_params:
            self.global_params[param_name] = new_value
            self.save_new_param_json()
            return True
        else:
            return False

    def change_dataset_param(self, param_name, new_value):
        if param_name in self.dataset_params:
            self.dataset_params[param_name] = new_value
            self.global_params["need_recalc_dataset_params"] = True
            self.save_new_param_json()
            return True
        else:
            return False

    def change_filter_param(self, param_name, new_value):
        if param_name in self.filter_params:
            self.filter_params[param_name] = new_value
            self.global_params["need_recalc_filter_params"] = True
            self.save_new_param_json()
            return True
        else:
            return False

    def change_box_param(self, param_name, new_value):
        if param_name in self.box_params:
            # if param_name == "total_num_spatial_boxes":
            #     assert (int(new_value**.5))**2 == new_value, "Please make sure Number of Spatial Boxes is a square number"
            self.box_params[param_name] = new_value
            self.global_params["need_recalc_box_params"] = True
            self.global_params["need_recalc_eigen_params"]=True
            self.global_params["need_recalc_roi_extraction_params"] = True
            self.save_new_param_json()
            return True
        else:
            return False

    def change_eigen_param(self, param_name, new_value):
        if param_name in self.eigen_params:
            self.eigen_params[param_name] = new_value
            self.global_params["need_recalc_eigen_params"] = True
            self.save_new_param_json()
            return True
        else:
            return False

    def change_roi_extraction_param(self, param_name, new_value):
        if param_name in self.roi_extraction_params:
            self.roi_extraction_params[param_name] = new_value
            self.global_params["need_recalc_roi_extraction_params"] = True
            self.save_new_param_json()
            return True
        else:
            return False

    def calculate_dataset(self)->np.ndarray:
        """Loads the dataset

        Returns
        -------
        """
        # TODO make it so it does't load the dataset every time

        self.dataset = load_filter_tif_stack(path=self.dataset_params[
            "dataset_path"], filter=False,
                                             median_filter=False,
                                             median_filter_size=(1,3,3),
                                             z_score=False, slice_stack=self.dataset_params[
                "slice_stack"],
                                             slice_start=self.dataset_params["slice_start"],

                                             slice_every=self.dataset_params["slice_every"])
        self.global_params["need_recalc_dataset_params"] = False
        self.shape = self.dataset.shape

        print("Finished Calculating Dataset")

        return self.dataset

    def calculate_filters(self):

        if self.global_params["need_recalc_filter_params"] or self.global_params["need_recalc_dataset_params"] or \
                not hasattr(self,"dataset_filtered"):
            dataset = self.calculate_dataset()
            self.dataset_filtered = filter_stack(stack=dataset,
                                                 median_filter_size=(1, self.filter_params["median_filter_size"], self.filter_params["median_filter_size"]),
                                                 median_filter=self.filter_params[
                                                     "median_filter"],
                                                 z_score=self.filter_params["z_score"])
            del dataset
            self.mean_image = np.mean(self.dataset_filtered, axis=0)
            self.max_image = np.max(self.dataset_filtered, axis=0)
            # self.temporal_correlation_image = calculate_temporal_correlation(self.dataset_filtered)
            self.global_params["need_recalc_filter_params"] = False
        return self.dataset_filtered






    def calculate_roi_extraction(self):
        if self.global_params["need_recalc_eigen_params"] or self.global_params[
            "need_recalc_roi_extraction_params"] or self.global_params[
            "need_recalc_box_parmas"] or self.global_params["need_recalc_dataset_params"] or \
                self.global_params["need_recalc_filter_params"]:
            assert (int(
                self.box_params[
                    "total_num_spatial_boxes"] ** .5)) ** 2 == self.box_params[
                                             "total_num_spatial_boxes"], "Please make sure Number of Spatial Boxes is a square number"
            try:
                self.clusters = process_data(num_threads=self.global_params[
                "num_threads"], test_images=False, test_output_dir="",
                                         save_dir=self.save_dir_path,
                                         save_intermediate_steps=self.global_params[
                     "save_intermediate_steps"],
                                         load_data=False, data_path="",
                                         image_data=self.calculate_filters(),
                                         eigen_vectors_already_generated=(not self.global_params[
                     "need_recalc_eigen_params"])and self.global_params["save_intermediate_steps"]and self.eigen_vectors_exist,
                                         save_embedding_images=True,
                                         total_num_time_steps=self.box_params["total_num_time_steps"],
                                         total_num_spatial_boxes=self.box_params[
                                             "total_num_spatial_boxes"],
                                         spatial_overlap=self.box_params["spatial_overlap"]//2, filter= False,
                                         median_filter=False,
                                         median_filter_size=(1,3,3),
                                         z_score=False, slice_stack=False,
                                         slice_every=1, slice_start=0, metric=self.eigen_params["metric"],
                                         knn=self.eigen_params["knn"],
                                         accuracy=self.eigen_params["accuracy"],
                                         connections=self.eigen_params["connections"],
                                         normalize_w_k=self.eigen_params["normalize_w_k"],
                                         num_eig=self.eigen_params["num_eig"],
                                         merge=self.roi_extraction_params["merge"],
                                         num_rois=self.roi_extraction_params["num_rois"],
                                         refinement=self.roi_extraction_params["refinement"],
                                         num_eigen_vector_select=self.roi_extraction_params[
                              "num_eigen_vector_select"],
                                         max_iter=self.roi_extraction_params["max_iter"],
                                         roi_size_min=self.roi_extraction_params["roi_size_min"],
                                         fill_holes=self.roi_extraction_params[
                                                 "fill_holes"],

                                         elbow_threshold_method=self.roi_extraction_params[
                     "elbow_threshold_method"],
                                         elbow_threshold_value=self.roi_extraction_params["elbow_threshold_value"],
                                         eigen_threshold_method=self.roi_extraction_params[
                     "eigen_threshold_method"],
                                         eigen_threshold_value=self.roi_extraction_params[
                     "eigen_threshold_value"],
                                         merge_temporal_coef=self.roi_extraction_params["merge_temporal_coef"],
                                         roi_size_max=self.roi_extraction_params["roi_size_max"])

                self.global_params["need_recalc_eigen_params"] = False
                self.save_rois(self.clusters)
                print("Calculating Time Traces:")
                self.time_traces = []
                for cluster in self.clusters:
                    # TODO maybe make this be a class variable
                    data_2d = reshape_to_2d_over_time(self.dataset_filtered)
                    time_trace = np.average(data_2d[cluster], axis=0)
                    self.time_traces.append(time_trace)
                self.gen_roi_display_variables()
                self.calculate_time_traces()

                self.rois_loaded = True
            except Exception as e:
                self.global_params["need_recalc_eigen_params"] = False
                logger1.error(e)
                print("Please try again there was an internal error in the roi extraction process")
                raise AssertionError()

    def gen_roi_display_variables(self):
        cluster_list_2d_cord = [
            pixel_num_to_2d_cord(x, volume_shape=self.dataset_filtered.shape) for x in
            self.clusters]
        self.cluster_max_cord_list = [np.max(x, axis=1) for x in cluster_list_2d_cord]
        self.cluster_min_cord_list = [np.min(x, axis=1) for x in cluster_list_2d_cord]
        self.pixel_with_rois_flat = np.zeros(
            [self.dataset_filtered.shape[1] * self.dataset_filtered.shape[2]])
        self.pixel_with_rois_color_flat = np.zeros(
            [self.dataset_filtered.shape[1] * self.dataset_filtered.shape[2], 3])
        for num, cluster in enumerate(self.clusters):
            cur_color = self.color_list[num % len(self.color_list)]
            self.pixel_with_rois_flat[cluster] = num + 1
            self.pixel_with_rois_color_flat[cluster] = cur_color
        edge_roi_image = feature.canny(np.reshape(self.pixel_with_rois_flat,
                                                [self.dataset_filtered.shape[1],
                                                 self.dataset_filtered.shape[2]]))
        self.edge_roi_image_flat = np.reshape(edge_roi_image,[-1,1]) *255
        self.pixel_with_rois_color = np.reshape(self.pixel_with_rois_color_flat,
                                                [self.dataset_filtered.shape[1],
                                                 self.dataset_filtered.shape[2], 3])
        try:
            self.eigen_norm_image = np.asarray(Image.open(os.path.join(self.save_dir_path, "embedding_norm_images/embedding_norm_image.png")))
        except:
            print("Can't generate eigen Norm image please try again")

    def calculate_time_traces(self):
        self.time_traces = [[]]*len(self.clusters)
        for cluster in range(len(self.clusters)):
            self.calculate_time_trace(cluster+1)

        if os.path.isdir(self.save_dir_path):
            pickle_save(self.time_traces, "time_traces", output_directory=self.save_dir_path)
        self.gen_roi_display_variables()
        self.rois_loaded = True
    def calculate_time_trace(self,roi_num):
        """
        Calculates a time trace for a certain ROI and save to time trace list
        Parameters
        ----------
        roi_num roi to calculate for

        Returns
        -------

        """
        cluster = self.clusters[roi_num-1]
        data_2d = reshape_to_2d_over_time(self.dataset_filtered)
        time_trace = np.average(data_2d[cluster], axis=0)
        self.time_traces[roi_num-1] = time_trace
        if os.path.isdir(self.save_dir_path):
            pickle_save(self.time_traces, "time_traces", output_directory=self.save_dir_path)
    def get_time_trace(self, num):
        return self.time_traces[num-1]








import numpy as np
from skimage import measure
from CIDAN.LSSC.functions.embeddings import embed_eigen_norm
from typing import Union, Any, List, Optional, cast, Tuple, Dict
from functools import reduce
from itertools import compress
from scipy.ndimage.morphology import binary_fill_holes
from matplotlib import pyplot as plt
from scipy.sparse.csgraph import connected_components as connected_components_graph
from scipy.sparse import csr_matrix
from dask import delayed
@delayed
def roi_extract_image(*, e_vectors: np.ndarray,
                      original_shape: tuple, original_2d_vol: np.ndarray, merge: bool,
                      num_rois: int, refinement: bool, num_eigen_vector_select: int,
                      max_iter: int, roi_size_min: int, fill_holes: bool,
                      elbow_threshold_method: bool, elbow_threshold_value: float,
                      eigen_threshold_method: bool,
                      eigen_threshold_value: float, merge_temporal_coef: float,
                      roi_size_limit: int, box_num:int) -> List[np.ndarray]:
    """
    Computes the Local Selective Spectral roi_extraction algorithm on an set of
    eigen vectors
    Parameters
    ----------
    e_vectors: Eigen vector in 2D numpy array
    original_shape: Original shape of image
    original_2d_vol: A flattened 2d volume of the original image, used for
        mergestep
    parameters: A parameter object, used params:
        num_rois: Number of rois
        refinement: If to do roi refinement
        num_eigen_vector_select: Number of eigen values to project into
        max_iter: Max amount of pixels to try and roi around
        roi_size_threshold: Min size for roi to be output if too big might
         limit number of rois
        fill_holes: fills holes in rois
        elbow: whether to use elbow thresholding in refinement step
        eigen_threshold_method: whether to use thresholding method when
            selecting eigen values
        eigen_threshold_value: value for said method
    Returns
    -------
    2D list of rois [[np.array of pixels roi 1], [
                          np.array  of pixels roi 2] ... ]
    It will have length num_rois unless max_iter amount is surpassed
    """
    print("Spatial Box {}: Starting ROI selection process".format(box_num))
    pixel_length = e_vectors.shape[0]
    pixel_embedings = embed_eigen_norm(
        e_vectors)  # embeds the pixels in the eigen space
    initial_pixel_list = np.flip(np.argsort(
        pixel_embedings))  # creates a list of pixels with the highest values
    # in the eigenspace this list is used to decide on the initial point
    # for the roi

    roi_list = []  # output list of rois

    # iter_counter is used to limit the ammount of pixels it tries
    # from initial_pixel_list
    iter_counter = 0
    while len(roi_list) < num_rois and len(
            initial_pixel_list) > 0 and iter_counter < max_iter:
        iter_counter += 1
        # print(iter_counter, len(roi_list),
        #       len(roi_list[-1]) if len(roi_list) > 0 else 0)
        initial_pixel = initial_pixel_list[0]  # Select initial point
        # select eigen vectors to project into
        small_eigen_vectors = select_eigen_vectors(e_vectors,
                                                   [initial_pixel],
                                                   num_eigen_vector_select,
                                                   threshold_method=eigen_threshold_method,
                                                   threshold=eigen_threshold_value)
        # print(small_eigen_vectors.shape)
        # TODO Find way to auto determin threshold value automatically max values
        # project into new eigen space
        small_pixel_embeding_norm = embed_eigen_norm(small_eigen_vectors)

        # calculate the distance between the initial point and each pixel
        # in the new eigen space
        small_pixel_distance = pixel_distance(small_eigen_vectors,
                                              initial_pixel)

        # selects pixels in roi
        pixels_in_roi = np.nonzero(
            small_pixel_distance <= small_pixel_embeding_norm)[0]

        # runs a connected component analysis around the initial point
        # in original image
        pixels_in_roi_comp = connected_component(pixel_length,
                                                     original_shape,
                                                     pixels_in_roi,
                                                     initial_pixel)
        pixels_in_roi_final = pixels_in_roi_comp

        # runs refinement step if enabled and if enough pixels in roi
        if refinement:  # TODO Look at this again
            # and len(pixels_in_roi_final) > \
            #  roi_size_threshold / 2
            # selects a new set of eigenvectors based on the pixels in roi
            rf_eigen_vectors = select_eigen_vectors(e_vectors,
                                                    pixels_in_roi_final,
                                                    num_eigen_vector_select,
                                                    threshold_method=eigen_threshold_method,
                                                    threshold=eigen_threshold_value)

            # embeds all pixels in this new eigen space
            rf_pixel_embedding_norm = embed_eigen_norm(rf_eigen_vectors)

            # selects the initial point based on the pixel with max in
            # the new embedding space
            rf_initial_point = rf_select_initial_point(rf_pixel_embedding_norm,
                                                       pixels_in_roi_final)

            # calculate the distance between the initial point and each pixel
            # in the new eigen space
            rf_pixel_distance = pixel_distance(rf_eigen_vectors,
                                               rf_initial_point)

            # selects pixels in roi
            if elbow_threshold_method:
                threshold = elbow_threshold_value * elbow_threshold(rf_pixel_distance,
                                                                     np.argsort(
                                                                         rf_pixel_distance),
                                                                     half=True,
                                                                     num=iter_counter)
                rf_pixels_in_roi = np.nonzero(
                    rf_pixel_distance < threshold)[0]
            else:
                rf_pixels_in_roi = np.nonzero(
                    rf_pixel_distance <= rf_pixel_embedding_norm)[0]

            # runs a connected component analysis around the initial point
            # in original image
            rf_pixels_in_roi_comp = connected_component(pixel_length,
                                                            original_shape,
                                                            rf_pixels_in_roi,
                                                            rf_initial_point)
            pixels_in_roi_final = rf_pixels_in_roi_comp

        # checks if roi is big enough
        # print("roi size:", len(pixels_in_roi_final))

        if roi_size_min < len(
                pixels_in_roi_final) < roi_size_limit:
            roi_list.append(pixels_in_roi_final)

            # takes all pixels in current roi out of initial_pixel_list
            initial_pixel_list = np.extract(
                np.in1d(initial_pixel_list, pixels_in_roi_final,
                        assume_unique=True, invert=True),
                initial_pixel_list)
            if initial_pixel not in pixels_in_roi_final:
                initial_pixel_list = np.delete(initial_pixel_list, 0)

            # print(len(initial_pixel_list))
        else:
            # takes current initial point and moves it to end of
            # initial_pixel_list
            initial_pixel_list = np.delete(
                np.append(initial_pixel_list, initial_pixel_list[0]), 0)
    if fill_holes:
        roi_list = fill_holes_func(roi_list, pixel_length, original_shape)
    # Merges rois
    if merge:
        roi_list = merge_rois(roi_list,
                                  temporal_coefficient=merge_temporal_coef,
                                  original_2d_vol=original_2d_vol)
        if fill_holes:
            roi_list = fill_holes_func(roi_list, pixel_length, original_shape)
    return roi_list


def fill_holes_func(roi_list: List[np.ndarray], pixel_length: int,
               original_shape: Tuple[int]) -> List[np.ndarray]:
    """
    Close holes in each roi
    Parameters
    ----------
    roi_list
    pixel_length
    original_shape

    Returns
    -------
    roi list with holes filled
    """
    for num, roi in enumerate(roi_list):
        original_zeros = np.zeros((pixel_length))
        original_zeros[roi] = 255
        image_filled = binary_fill_holes(np.reshape(original_zeros,
                                                    original_shape[1:]))
        image_filled_2d = np.reshape(image_filled, (-1))
        roi_list[num] = np.nonzero(image_filled_2d)[0]
    return roi_list


def pixel_distance(eigen_vectors: np.ndarray, pixel_num: int) -> np.ndarray:
    """
    Calculates squared distance between pixels in embedding space and initial_point
    Parameters
    ----------
    eigen_vectors: The eigen vectors describing the vector space with
        dimensions number of pixels in image by number of eigen vectors
    pixel_num: The number of the initial pixel in the eigen vectors

    Returns
    -------
    A np array with dim: number of pixels in image
    """
    return np.sum(np.power(eigen_vectors - eigen_vectors[pixel_num], 2),
                  axis=1)


def connected_component(pixel_length: int, original_shape: Tuple[int],
                        pixels_in_roi: np.ndarray,
                        initial_pixel_number: int) -> np.ndarray:
    """
    Runs a connected component analysis on a group of pixels in an image
    Parameters
    ----------
    pixel_length: Number of pixels in image
    original_shape: Ariginal shape of image
    pixels_in_roi: A tuple with the first entry as a
    initial_pixel_number: The number of the original pixel in the
        flattened image


    Returns
    -------
    An subset of the original pixels that are connected to initial pixel
    """
    # TODO add in im fill before connected component
    # first creates an image with pixel values of 1 if pixel in roi
    original_zeros = np.zeros(pixel_length)
    original_zeros[pixels_in_roi] = 1
    pixel_image = np.reshape(original_zeros, original_shape[1:])

    # runs connected component analysis on image
    blobs_labels = np.reshape(measure.label(pixel_image, background=0),
                              (-1))
    correct_label = blobs_labels[initial_pixel_number]

    # filters pixels to only ones with same label as initial pixel
    pixels_in_roi_new = np.nonzero(
        blobs_labels == correct_label)[0]
    return pixels_in_roi_new


def select_eigen_vectors(eigen_vectors: np.ndarray,
                         pixels_in_roi: np.ndarray,
                         num_eigen_vector_select: int,
                         threshold_method: bool = False,
                         threshold: float = .1) -> np.ndarray:
    """
    Selects eigen vectors that are most descriptive of a set a points
    Parameters
    ----------
    eigen_vectors: The eigen vectors describing the vector space with
        dimensions number of pixels in image by number of eigen vectors
    pixels_in_roi: Np array of indices of all pixels in roi
    num_eigen_vector_select: Number of eigen vectors to select
    threshold_method: this is a bool on whether to run the threshold method to select the eigen vectors
    threshold

    Returns
    -------
    the eigen vectors describing the new vector space with
        dimensions number of pixels in image by numb_eigen_vector_select

    """
    pixel_eigen_vec_values = np.abs(np.sum(eigen_vectors[pixels_in_roi],
                                           axis=0))  # TODO DONE add in absolute value for only the non-refinement step and add absolute value of the sum
    pixel_eigen_vec_values_sort_indices = np.flip(
        np.argsort(pixel_eigen_vec_values))
    if threshold_method:

        threshold_filter = pixel_eigen_vec_values > threshold * \
                           pixel_eigen_vec_values[
                               pixel_eigen_vec_values_sort_indices[0]]
        small_eigen_vectors = eigen_vectors[:, np.nonzero(threshold_filter)[0]]

    else:
        pixel_eigen_vec_values_sort_indices = np.flip(
            np.argsort(
                pixel_eigen_vec_values))  # TODO DONE add thresholding method to this too .1 is a good threshold
        small_eigen_vectors = eigen_vectors[:,
                              pixel_eigen_vec_values_sort_indices[
                              :num_eigen_vector_select]]
    return small_eigen_vectors


def rf_select_initial_point(pixel_embedings: np.ndarray,
                            pixels_in_roi: np.ndarray):
    """
    Selects an initial point for roi_extraction based on the pixels in current
    roi, this is part of the refinement step
    Parameters
    ----------
    pixel_embedings: The embedings of each pixel in a vector space
    pixels_in_roi: a list of the indices of the pixel in the roi

    Returns
    -------
    an indice for the initial point pixel
    """
    indice_in_roi = \
        np.flip(np.argsort(pixel_embedings[pixels_in_roi]))[0]
    return np.sort(pixels_in_roi)[indice_in_roi]


def elbow_threshold(pixel_vals: np.ndarray, pixel_val_sort_indices: np.ndarray,
                    half: bool = True, num=0) -> float:
    """
    Calculates the elbow threshold for the refinement step in roi_extraction
    algorithm, determines the pixel that is farthest away from line drawn
    from first to last or first to middle pixel

    Projects each point(pixel # in sorted pixel_val, distance val) to the line then subtracts that from the points value to find distance
    Parameters
    ----------
    pixel_vals: The distance values for each pixel
    pixel_val_sort_indices: The array necessary to sort said pixels from lowest
     to highest
    half: whether to run only with the closest half of the pixels, recommended

    Returns
    -------
    float, the optimal threshold based on elbow
    """
    n_points = len(pixel_vals) if not half else len(pixel_vals) // 2
    # fig, ax = plt.subplots(figsize=(4, 4))
    # ax.scatter(list(range(n_points)),
    #            pixel_vals[pixel_val_sort_indices[:n_points]])
    # fig.savefig(
    #     "/data2/Sam/pythonTestEnviroment/output_images/plots/dist_plot_{}1.png".format(num),
    #     aspect='auto')
    # plt.close('all')
    pixel_vals_sorted_zipped = np.array(list(
        zip(range(n_points), pixel_vals[pixel_val_sort_indices[:n_points]])))

    first_point = pixel_vals_sorted_zipped[0, :]
    last_point = pixel_vals_sorted_zipped[-1, :]
    line_vec = last_point - first_point

    line_vec_norm = line_vec / (np.sum(np.power(line_vec, 2)) ** .5)
    dist_from_first = pixel_vals_sorted_zipped - first_point
    proj_point_to_line = np.matmul(dist_from_first,
                                   line_vec_norm[:, None]) * line_vec_norm
    vec_to_line = dist_from_first - proj_point_to_line
    dist_to_line = np.power(np.sum(np.power(vec_to_line, 2), axis=1), .5)
    # fig, ax = plt.subplots(figsize=(4, 4))
    # ax.scatter(list(range(n_points)),
    #            dist_from_first[:n_points,1])
    # fig.savefig(
    #     "/Users/sschickler/Documents/LSSC-python/output_images/plots/dist_plot_{}2.png".format(num),
    #     aspect='auto')
    dist_max_indice = np.argmax(dist_to_line)
    threshold = pixel_vals_sorted_zipped[dist_max_indice][1]
    return threshold


def merge_rois(roi_list: List,
               temporal_coefficient: float, original_2d_vol: np.ndarray):
    # TODO is this the most efficient implementation I can do
    """
    Merges rois based on temporal and spacial overlap
    Parameters
    ----------
    eigen_vectors
    roi_list
    temporal_coefficient

    Returns
    -------

    """
    A = np.zeros([original_2d_vol.shape[0], len(roi_list)], dtype=bool)
    for num, cluster in enumerate(roi_list):
        A[cluster,num] = True
    A_graph = np.matmul(A.transpose(),A)
    A_csr = csr_matrix(A_graph)
    connected = connected_components_graph(A_csr,False,return_labels=True)
    roi_groups = [[] for _ in range(len(roi_list))]
    for num in range(len(roi_list)):
        roi_groups[connected[1][num]].append(roi_list[num])

    new_rois = []
    for group in roi_groups:
        group_zipped = list(enumerate(group))
        timetraces = [np.mean(original_2d_vol[roi], axis=0) for roi in group]
        while len(group_zipped)>0:
            first_num, first_roi = group_zipped.pop(0)
            rois_to_merge = []
            for num, roi in enumerate(group_zipped):
                if compare_time_traces(timetraces[first_num],timetraces[roi[0]])>temporal_coefficient:
                    rois_to_merge.append(num)
            first_roi = list(reduce(combine_rois, [first_roi]+[group_zipped[x][1] for x in rois_to_merge]))
            for num in rois_to_merge[::-1]:
                group_zipped.pop(num)
            new_rois.append(first_roi)

    return new_rois
def compare_time_traces(trace_1, trace_2):
    """
    Compares two timetraces based on person correlation
    Parameters
    ----------
    trace_1
    trace_2

    Returns
    -------
    the correlation
    """
    trace_1_mean = np.mean(trace_1)
    trace_2_mean = np.mean(trace_2)
    trace_1_sub_mean = (trace_1-trace_1_mean)
    trace_2_sub_mean = (trace_2-trace_2_mean)
    top = np.dot(trace_1_sub_mean, trace_2_sub_mean)
    bottom = (np.dot(trace_1_sub_mean, trace_1_sub_mean)*
              np.dot(trace_2_sub_mean,trace_2_sub_mean))**.5
    return top/bottom

def compare_roi(roi1: List[int],
                    roi2: List[int], temporal_coefficient: int,
                    original_2d_vol: np.ndarray) -> bool:
    """
    Compares two rois and sees if they meet the standards for being combined #calculate average time trace given spacial support correlation is above a certain factor
    Parameters
    ----------
    roi1
    roi2
    temporal_coefficient
    original_2d_vol

    Returns
    -------
    """

    if any([x in roi2 for x in roi1]):  # check spacial simlarity
        roi1_time_trace_avg = np.mean(original_2d_vol[roi1],
                                          axis=0)  # TODO Do I zscore the time trace
        roi2_time_trace_avg = np.mean(original_2d_vol[roi2], axis=0)
        roi1_time_trace_avg_z_scored = (roi1_time_trace_avg - np.mean(
            roi1_time_trace_avg)) / np.std(roi1_time_trace_avg)
        roi2_time_trace_avg_z_scored = (roi2_time_trace_avg - np.mean(
            roi2_time_trace_avg)) / np.std(roi2_time_trace_avg)
        if np.linalg.norm(
                roi1_time_trace_avg_z_scored -
                roi2_time_trace_avg_z_scored) < temporal_coefficient * (
                original_2d_vol.shape[0] ** .5):
            return True

    return False


def combine_rois(roi1: List[int], roi2: List[int]) -> List[int]:
    """
    Combines two lists of clusters into one
    Parameters
    ----------
    roi1 One list of pixels in cluster each pixel # is based on 1d representation of
        image
    roi2 List for other ROI

    Returns
    -------
    List of merged ROI
    """
    roi2_not_in_1 = list(filter(lambda x: x not in roi1, roi2))
    return np.array(list(roi1) + roi2_not_in_1)
